// -------------------------------------------------------------------------\\
//          
// File:        target.h
// Author:      Bettina McCormick
//          
// Project:     Atlantis: COP 2224 VCC; Summer 1999
//          
// Created:     June 05, 1999
// Updated:     July 25, 1999
// Date due:    August 04, 1999
//             
// Description: Specification for Target base class
//
//              Base class for inheritance- contains virtual methods that 
//              must be implemented by subclasses that inherit from it
//        
// Notes:       Turn on symbol TARGET_TEST to test this component
//
// -------------------------------------------------------------------------//
    
#ifndef TARGET_H  // standard header wrapper
#define TARGET_H
    
#include <iostream>
#include <string>
    
#include "constuff.h"
    
namespace Atlantis  // project namespace
{
  
  enum TargetStatus   // for better readability
  {
    UNKNOWN,    // 0
    NO_DAMAGE,  // 1
    DAMAGED,    // 2
    DESTROYED,  // 3
    STATUS_MAX  // 4
  };
    
  enum TargetType   // for better readability
  {
    POD,      // 0
    CRYSTAL   // 1
  };
    
  // static variables - belong to namespace; can be access "globally"
  static const std::string StatusString[] = 
  {                                 
      "Unknown",
      "No Damage",
      "Damaged",
      "Destroyed"
  }; // const array of strings for use in display
    
  static const std::string TargetTypeString[] = 
  {                                 
      "Evil Pod",
      "Crystal"
  }; // const array of strings for use in display 
      
      
  class Target
  {
  protected:  // can be accessed by this and subclasses
    enum CONSTANTS { MIN_ARMOR = 1, MAX_ARMOR = 5, MAX_DEPTH = 10 };
    
  public:
    Target( std::string n ="Uninitialized" ,int a = MAX_ARMOR, 
            int d = MAX_DEPTH, int v = MAX_ARMOR * 100, int t = POD );
    /**------------------------------------------------------------------------
    // public constructor: takes string name, armor level, depth limit & type
    //-----------------------------------------------------------------------*/
      
    virtual ~Target();
    /**------------------------------------------------------------------------
    // destructor: virtual
    //-----------------------------------------------------------------------*/
    
    TargetStatus Get_status( void ) const;
    /**------------------------------------------------------------------------
    // accessor method: returns status of target
    //-----------------------------------------------------------------------*/
        
    int Get_depth_limit ( void ) const;
    /**------------------------------------------------------------------------
    // accessor method: returns depth limit of target
    //-----------------------------------------------------------------------*/
    
    int Get_armor ( void ) const;
    /**------------------------------------------------------------------------
    // accessor method: returns armor level of target 
    //-----------------------------------------------------------------------*/
    
    int Get_value( void ) const;
    /**------------------------------------------------------------------------
    // returns value of target
    //-----------------------------------------------------------------------*/
        
    int Get_type ( void )  const;
    /**------------------------------------------------------------------------
    // accessor method: returns type of target
    //-----------------------------------------------------------------------*/
      
    void Set_inLostcity( bool ilc ); 
    /**------------------------------------------------------------------------
    // mutator method: sets bool data inLostcity, true or false
    //-----------------------------------------------------------------------*/
    
    void Reset( void ); 
    /**------------------------------------------------------------------------
    // mutator method: resets data to new target status
    //-----------------------------------------------------------------------*/
    
    // -----------------------------------Method interface for derived objects
    virtual bool Hit( void );
    /**------------------------------------------------------------------------
    // hits target; updates status; returns true if target is alive and hitable
    //-----------------------------------------------------------------------*/
    
    virtual void Show( void ) const;   
    /**------------------------------------------------------------------------
    // shows the targets name, armor level, hits & status
    //-----------------------------------------------------------------------*/
    
    virtual void Show ( bool is_hit ) const;
    /**------------------------------------------------------------------------
    // overloaded:show method: is_hit is true if pod is hit, false if missed
    //-----------------------------------------------------------------------*/
    
    #ifndef TARGET_TEST
    virtual void Abstract ( void ) = 0;
    /**------------------------------------------------------------------------
    // pure virtual method creates abstract class for deployed application;
    // cannot create an instance of this class outside the test driver.
    //-----------------------------------------------------------------------*/
    #endif 
    
  protected: //------------------------------------------protected data members
    TargetStatus status;  // status of target
    std::string name;     // name of target
    int  armor;           // armor protection
    int  hits;            // hits it has taken
    int  value;           // value to player's score if destroyed
    int  depth_limit;     // depth_limit position
    int  target_type;     // type crystal or pod
    bool inLostcity;      // true if placed in Lostcity
    
  private:
    Target( const Target & t );
    /**------------------------------------------------------------------------
    // copy constructor: takes reference to a Target
    // private prevents passing target by value 
    //-----------------------------------------------------------------------*/
    
  };
  
} // namespace Atlantis
    
#endif // TARGET_H
    
// eof target.h

