// -------------------------------------------------------------------------\\
//          
// File:        constuff.cpp
// Author:      Bettina McCormick
//          
// Project:     Atlantis: COP 2224 VCC; Summer 1999
//          
// Created:     June 01, 1999
// Updated:     July 31, 1999
// Date due:    August 04, 1999
//             
// Description: Implementation of Constuff class
//
//              Static class - methods can be accessed without an instance 
//              of the class.
//              This class contains many handy methods for use in console 
//              applications  
//    
// Notes:       Turn on symbol CONSTUFF_TEST to test this component
//    
// -------------------------------------------------------------------------//
    
#include <strstream>
#include <cstdlib>
#include <cstdio>
    
#include "constuff.h"
    
namespace Atlantis // project namespace
{
  
  std::string 
  Constuff::Int2String( int i )
  /**------------------------------------------------------------------------
  // converts int to ascii
  //-----------------------------------------------------------------------*/
  {      
    std::strstream stream;
    std::string str;
    stream << i;
    stream >> str;

    return str;
  } // end Int2String( int i )
      
       
  std::string
  Constuff::Double2String( double d )
  /**------------------------------------------------------------------------
  // converts float to ascii
  //-----------------------------------------------------------------------*/
  {
    std::strstream stream;
    std::string str;
    stream << d;
    stream >> str;

    return str;
  } // end Double2String( double d )
  
  
  void 
  Constuff::UserEntry ( std::string label, int & entry, int min, int max )
  /**------------------------------------------------------------------------
  // overloaded method: takes a string, ref to int and 2 ints
  // outputs label to screen with range to choose from
  // prompts user for entry between range of min and max, inclusive
  // memory in entry is changed to user entry from keyboard
  // required that min not less than 0
  //-----------------------------------------------------------------------*/
  {
    const int MAX_LENGTH = 15; // maximum length read in from keyboard

    char buff[MAXBUFF] = {0}; // variable to hold input from keyboard

    bool test = false;        // test for return key error

    bool valid = true;  // assume valid entry

    entry = -1;  // set entry to value out of range

    std::string range = " <" + Int2String(min) + " to "   // range string
                             + Int2String(max) + ">";
    
   // std::cin.ignore(MAXBUFF,'\n');   // clear anything in buffer    
    
    do  // output to screen at least once
    {
      valid = true; // assume true each try
      std::cout << "\nEnter " << label  << range << ": ";  // prompt user
      std::cin.get( buff, MAX_LENGTH );  // get user entry
      std::cin.ignore( MAXBUFF,'\n' );  // clear anything in buffer
      test = false;
      for ( int i=0; i < strlen( buff ); ++i )
      { 
        test = true;
        if ( !isdigit( buff[i] ) ) // if not a digit, input is not valid
        {
          valid = false;
        }
      }
      if (test == false )
      {
        valid = false;
        strcpy(buff,"Return Key");
        entry = -1;
        std::cin.clear();
        ReturnError();
      }
      if ( valid == true )   // if valid remained true
      {
        entry = atoi( buff );  // convert characters to int
        if ( entry < min || entry > max )  // check int is in range
        {
          std::cout << "\n\t" << entry << " is out of range"<< range 
                    << std::endl;
        }
      }
      else // if not valid entry, prompt error
      {
        std::cout << "\n\t\"" << buff << "\" is an INVALID ENTRY:" 
                  << std::endl;
      }
    } while ( entry < min || entry > max );
    
  } // end UserEntry ( std::string label, int & entry, int min, int max )
      
      
  void 
  Constuff::UserEntry ( std::string label, std::string & entry, int length )
  /**------------------------------------------------------------------------
  // overloaded method: takes a string, ref to string and an int
  // prompts user for a string
  // memory in entry is changed to user entry from keyboard
  // length is limit of string length read from keyboard
  //-----------------------------------------------------------------------*/
  {
    char buff[MAXBUFF] = {0};  

    std::cout << "Enter " << label  << ": "; // prompts user for string
    std::cin.getline(buff, length);// bug in VCPP 5.0 doesn't work correctly
    entry = buff;  // set entry to string from user
  } // end UserEntry ( std::string label, std::string & entry, int length )
  
    
  void 
  Constuff::UserEntry ( std::string label, std::string & entry )
  /**------------------------------------------------------------------------
  // overloaded method: takes a string, ref to string
  // prompts user for a string
  // memory in entry is changed to user entry from keyboard
  // uses default length for string of 40
  //-----------------------------------------------------------------------*/
  {
    const int MAX_LENGTH = 40; // max amount read in from keyboard
    char buff[MAXBUFF] = {0};  

    std::cout << "Enter " << label  << ": "; // prompt user for string
    std::cin.getline(buff, MAX_LENGTH);// bug in VCPP 5.0: fails
    entry = buff;  // set entry to string from user
  } // end
    
    
  void
  Constuff::PauseScreen( void )
  /**------------------------------------------------------------------------
  // pauses the screen prompting user to press any key to continue
  //-----------------------------------------------------------------------*/
  {
    char buff[1] = {0};
    std::cout << "\nPress \"Enter\" to Continue ";
    gets(buff);
  } // end PauseScreen( void )
      
      
  void
  Constuff::ReturnError( void )
  /**------------------------------------------------------------------------
  // fixes return keyy error
  //-----------------------------------------------------------------------*/
  {
    char buff[1] = {0};
    std::cout << "\n\t Error:";
    gets(buff);
  } // end PauseScreen( void )

  void 
  Constuff::ClearScreen( void )
  /**------------------------------------------------------------------------
  // clears the screen visible to the user on the console
  //-----------------------------------------------------------------------*/
  {
    std::cout << "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n" 
      << std::endl;
  } // end ClearScreen( void )
  
  
  void  
  Constuff::Welcome( void ) 
  /**------------------------------------------------------------------------
  // displays a description of the program to the user 
  // change std::cout lines for each purpose
  //-----------------------------------------------------------------------*/
  {
    std::cout << "Welcome!  \n\t"
      
      
      
              << std::endl;
    
  } // end Welcome( void ) 
  
  
  bool 
  Constuff::GetChoice( void ) 
  /**------------------------------------------------------------------------
  // queries the user for a choice of yes (true) or no (false)
  // returns their choice
  //-----------------------------------------------------------------------*/
  {
    bool  choice = false;// boolean to be determined and returned
    char  temp[1] = {0}; // temp variable to hold input from keyboard
    
    std::cout << "(1 = YES / 2 = NO):  ";
    std::cin  >> temp;    // grab anything else they might have entered
    
    while ( (temp[0] != '1') && (temp[0] != '2') )
    {  // guarentees the only values entered are 1 or 2   
      std::cout << "\n\tYou must enter either  1 = YES  or  2 = NO:  ";
      std::cin >> temp;
    }
    
    if (temp[0] == '1')
    {
      choice = true; 
    }
    
    return ( choice );
    
  } // end GetChoice( void )  
    
      
  int 
  Constuff::GetRandNum( int max_rand )
  /**------------------------------------------------------------------------
  // returns a random number between 1 and the max_rand inclusive
  //-----------------------------------------------------------------------*/
  {
    int randnum = 0;

    randnum = rand() % max_rand;  // get random number

    while ( randnum == 0 )
    { // logic to prevent randnum from being zero
      if ( max_rand == 1 )
      {
        randnum = max_rand;
      }
      else
      {
        randnum = rand() % max_rand;  // get random number
      }
    } // end while == 0

    return ( randnum );
  } // end GetRandNum( int max_rand )
  
} // namespace Atlantis
      
        
#ifdef CONSTUFF_TEST
/**------------------------------------------------------------------------
// unit test for the methods in this class
//-----------------------------------------------------------------------*/
    
int
main ( void )
{     
  std::cout << "Constuff Unit Test\n" << std::endl;
  
  const int STRLENGTH = 50;   // consts for readability
  const int MIN_RANGE = 1;        
  const int MAX_RANGE = 10;
    
  std::string label1 = "a number";  // generic labels
  std::string label2 = "a name";
  std::string label3;
  int myentry= 0;
  int myrand = 0;
  int i = 0;
    
    
  Atlantis::Constuff::Welcome();
    
  std::cout << "\nWould you like to use this program?  ";
  if ( Atlantis::Constuff::GetChoice() == true )
  {
    std::cout << "\n\tYou have chosen YES" << std::endl;
  }
  else
  {
    std::cout << "You have chosen NO" << std::endl;
  }
  
  // test first overloaded UserEntry
  Atlantis::Constuff::UserEntry(label1, myentry, MIN_RANGE, MAX_RANGE);
  std::cout << "\n\tYou entered:  " << myentry  << std::endl << std::endl;
    // test second overloaded UserEntry
  Atlantis::Constuff::UserEntry(label2, label3, STRLENGTH);
  std::cout << "\n\tYou entered:  " << label3  << std::endl << std::endl;
  // test third overloaded UserEntry
  Atlantis::Constuff::UserEntry(label2, label3);
  std::cout << "\n\tYou entered:  " << label3  << std::endl << std::endl;
    
  Atlantis::Constuff::PauseScreen();
  Atlantis::Constuff::ClearScreen();

  // test GetRandnum
  srand( time ( NULL ) );  // seed rand with time
  for (i = 1; i <= MAX_RANGE; i++)
  {
    myrand = Atlantis::Constuff::GetRandNum( i );
    std::cout << "\tYour Random # is:  " << myrand  << std::endl << std::endl;
  }
  Atlantis::Constuff::PauseScreen();
  Atlantis::Constuff::ClearScreen();
    
  for (i = 1; i <= MAX_RANGE; i++)
  {
    myrand = Atlantis::Constuff::GetRandNum( i );
    std::cout << "\tYour Random # is:  " << myrand  << std::endl << std::endl;
  }
  
  Atlantis::Constuff::PauseScreen();

  return (0);
  
} // end main
    
#endif // CONSTUFF_TEST
    
// eof constuff.cpp

