// -------------------------------------------------------------------------\\
//          
// File:        crystal.cpp
// Author:      Bettina McCormick
//          
// Project:     Atlantis: COP 2224 VCC; Summer 1999
//          
// Created:     July 02, 1999
// Updated:     July 28, 1999
// Date due:    August 04, 1999
//             
// Description: Implementation for Crystal class
//
//              Inherits from Target - Crystal "is a" Target
//        
// Notes:       Turn on symbol CRYSTAL_TEST to test the component
//      
// -------------------------------------------------------------------------//
      
#include "crystal.h"      // specification
      
namespace Atlantis    // project namespace
{  
  
  static const std::string CrystalTypeString[] = 
  {                                 
      "Laughter",
      "Forgiveness",
      "Peace",
      "Joy",
      "Love"
  }; // const array of strings for use in display
    
    
  Crystal *
  Crystal::Create( int crystal_type )
  /**------------------------------------------------------------------------
  // no public constructor: this calls private constructor to create instance
  // of crystal and returns crystal *
  //-----------------------------------------------------------------------*/
  {
    std::string crystalname;  // temp values mirroring the data of a target
    int d = 0;            // to be passed to target constructor
    int a = 0;
    int v = 0;
    int t = CRYSTAL;

    switch ( crystal_type ) 
    {                     // set values according to the crystal type
    case LAUGHTER:
      crystalname = CrystalTypeString[LAUGHTER];// least threatening crystal
      d = Constuff::GetRandNum(MAX_DEPTH);  // get random depth level
      a = MIN_ARMOR;                        // least armor protection
      v = a * 150;                  // value is multiple of armor
      break;
    case FORGIVENESS:
      crystalname = CrystalTypeString[FORGIVENESS]; // crystals increase in size
      d = Constuff::GetRandNum(MAX_DEPTH);
      a = MIN_ARMOR + 1;
      v = a * 100;  // multiple of armor varies - either 150 or 100
      break;
    case PEACE:
      crystalname = CrystalTypeString[PEACE];
      d = Constuff::GetRandNum(MAX_DEPTH);  // max depth possible varies
      a = MIN_ARMOR + 2;
      v = a * 150;                  // alternate multiple value
      break;
    case JOY:
      crystalname = CrystalTypeString[JOY];
      d = Constuff::GetRandNum(MAX_DEPTH);
      a = MIN_ARMOR + 3;
      v = a * 100;
      break;
    case LOVE:
      crystalname = CrystalTypeString[LOVE];   // strongest Crystal
      d = Constuff::GetRandNum(MAX_DEPTH);     // larget maximum depth
      a = MIN_ARMOR + 4;                // most armor protection
      v = a * 150;                      // largest value
      break;  
    default:
      crystalname = CrystalTypeString[LOVE];
      d = Constuff::GetRandNum(10);     // default is strongest Crystal
      a = MIN_ARMOR + 4;
      v = a * 150;
      break;
    };
    
    Crystal * newCrystal = new Crystal ( crystalname, a, d, v, t ); 
                              // call constructor
    return( newCrystal );         // return Crystal
  } // end Create( int crystal_type )
    
    
  Crystal::Crystal( std::string name, int armor, int maxdepth, int val,
                    int t_type ):Target( name, armor, maxdepth, val, t_type )
  /**------------------------------------------------------------------------
  // private constructor calls virtual constructor of base class
  //-----------------------------------------------------------------------*/
  {
    #ifdef CRYSTAL_TEST
    std::cout << "Constructed Crystal " << name << std::endl;
    #endif
  } // end Crystal( std::string name, int a, int d, int t )
    
    
  Crystal::~Crystal()
  /**------------------------------------------------------------------------
  // destructor
  //-----------------------------------------------------------------------*/
  {
    // no dynamic memory created in constructor 
    // this will be called when pointer to Crystal is deleted
    #ifdef CRYSTAL_TEST // only for testing
    std::cout << "Client code has successfully Destructed Crystal " 
              << name << std::endl;
    #endif
  } // end destructor
  
  
  void 
  Crystal::Abstract ( void ) 
  /**------------------------------------------------------------------------
  // fullfills virtual bass class method
  //-----------------------------------------------------------------------*/
  { 
    /* no body */ 
  }
      
    
  bool
  Crystal::Hit ( void )
  /**------------------------------------------------------------------------
  // returns true and hits Crystal if not already destroyed, false otherwise, 
  //-----------------------------------------------------------------------*/
  {
    bool hit_status = Target::Hit(); // true if taget is hitable
  
    return( hit_status );
  } // end Hit ( void )
    
      
  void
  Crystal::Show ( void ) const
  /**------------------------------------------------------------------------
  // calls base class method: shows crystals name, armor level, hits & status
  //-----------------------------------------------------------------------*/
  {
    Target::Show();
  } // end Show ( void ) const
    
    
  void
  Crystal::Show ( bool hitflag ) const
  /**------------------------------------------------------------------------
  // overloaded show method: hitflag is true if crystal is hit, false if missed
  //-----------------------------------------------------------------------*/
  {   
    Target::Show( hitflag );    // calls base class method
  } // end Show ( bool hitflag ) const
  
} // namespace Atlantis
    
    
#ifdef CRYSTAL_TEST
  /**------------------------------------------------------------------------
  // unit test for the methods in this class
  //-----------------------------------------------------------------------*/
  
int
main ( void )
{
  std::cout << "Crystal Unit Test\n" << std::endl;
      
  bool hit = true;
  int type = 0;
  // declare pointer to crystal
  Atlantis::Crystal * my_crystal = Atlantis::Crystal::Create( Atlantis::PEACE );
  std::cout << std::endl;

  type = my_crystal->Get_type();       // show target type
  std::cout << "Target is " << Atlantis::TargetTypeString [ type ] 
            << "\nValue of target is " << my_crystal->Get_value() << std::endl;
  std::cout << std::endl;
    
  hit = my_crystal->Hit();                           // hit and show
  my_crystal->Show( hit );                           // 1 damage
  Atlantis::Constuff::PauseScreen();  
  Atlantis::Constuff::ClearScreen(); 
  hit = my_crystal->Hit();                           // hit and show
  my_crystal->Show( hit );                           // 2 damage
  Atlantis::Constuff::PauseScreen();  
  Atlantis::Constuff::ClearScreen(); 
  hit = my_crystal->Hit();                           // hit and show
  my_crystal->Show( hit );                           // 3 destroyed 
  Atlantis::Constuff::PauseScreen();  
  Atlantis::Constuff::ClearScreen(); 
  hit = my_crystal->Hit();                           // hit and show
  my_crystal->Show( hit );                           // destroyed  
  Atlantis::Constuff::PauseScreen();  
  Atlantis::Constuff::ClearScreen(); 
  hit = my_crystal->Hit();                           // hit and show
  my_crystal->Show( hit );                           // already destroyed  
  Atlantis::Constuff::PauseScreen();

  std::cout << "\nResetting target:" << std::endl;
  my_crystal->Reset();
  my_crystal->Show();

  if ( my_crystal )    // check pointer
  {
    delete my_crystal; // delete crystal
    my_crystal = 0;    // set pointer null
  }
  
  return (0);
} // end main
    
#endif // CRYSTAL_TEST

// eof crystal.cpp

