// ***************************************************************************
//    
// File:        lmach.cpp
// Author:      Bettina McCormick
//    
// Project:     Midterm: COP 2224 VCC; Summer 1999
// Template:    Created by Professor Richard Mudge
//    
// Created:     June 17, 1999
// Updated:     June 22, 1999
// Date due:    June 28, 1999
//       
// Description: Implementation for the LottoMachine object that is used with  
//              the ball object in a simulation of a lotto machine.
//    
// ***************************************************************************
      
   // define LMACH_TEST in preprocessor definitions to run unit test            
      
   #include "lmach.h"
      
   static int SMALL = 0;  // initialized for small machine size logic later   
      
   LottoMachine::LottoMachine( unsigned char number_balls )
   //------------------------------------------------------------------------
   // initializes the object
   // sets all ball numbers and their picked state to false
   //------------------------------------------------------------------------
   {
      int i = 0;
         
      srand( time ( NULL ) );  // seed rand with time 
      
      machine_size = number_balls;
      
      if ( machine_size > ZERO_BASED )
      {  // for fantasy5 and lotto machines
         for (i = 0; i < machine_size; i++)   // set contents
         {                                    // starting at one
            contents[i].Set_number ( (unsigned char) i + 1 );
         } // end for set starting at one
      }
      else 
      {  // for cash3 and play4 machines
         SMALL++; // increment static variable
         for (i = 0; i < machine_size; i++)  // set contents
         {                                   // starting at zero
            contents[i].Set_number ( (unsigned char) i );
         } // end for set starting at zero
      }
      
      balls_picked = 0;
      
      for ( i = 0; i < machine_size; i++ )  // initialize indexes_tried
      {
         indexes_tried[i] = 0;
      }
      
      #ifdef LMACH_TEST // comment to output
      cout << "Constructor: created an instance of LottoMachine - size " 
           << static_cast<int>( machine_size ) << endl;
      #endif
      #ifdef PLAYSLIP_TEST // comment to output
      cout << "Constructor: created an instance of LottoMachine - size " 
           << static_cast<int>( machine_size ) << endl;
      #endif
      
   } // end Constructor
         
         
   LottoMachine::~LottoMachine()
   //------------------------------------------------------------------------
   // cleans up the object
   //------------------------------------------------------------------------
   {
      // no dynamic memory to deallocate
      #ifdef LMACH_TEST // comment to output
      cout << "Destructor: LottoMachine object has dropped out of scope" 
           << endl; 
      #endif
      #ifdef PLAYSLIP_TEST // comment to output
      cout << "Destructor: LottoMachine object has dropped out of scope\n" 
           << endl; 
      #endif
   } // end Destructor
         
         
   const Ball &  
   LottoMachine::Pick ( void )
   //------------------------------------------------------------------------
   // picks a ball and returns it
   //------------------------------------------------------------------------
   {
      static const int TRIED = 1;  // 1 means index has tried to be picked
      
      int index = 0, i = 0, test = 0;  // for incrementation
      int saftey_counter = 0;  // to kill while loop if all balls picked
   
      if ( (SMALL > 1) && (SMALL < 6) ) // if not 1st machine of this size
      {     // for small machines vary the first index
         SMALL++;
         index = rand() % machine_size;  // get random index
         test = index;
         while ( test == index )
         { // logic to vary index different than other machine of same size
            test = rand() % machine_size;  // get random index
            i++;
            if ( i == 10 )
            { // if loop 10 times, enough...get out
               break;
            }
         }
         index = test;
      } // end if not 1st machine of this size
      else
      { // else just pick a rand index
         index = rand() % machine_size;  // get random index
      } // end else

      while ( contents[index].Pick() == false ) // while ball already picked
      {
      // get another random index
         while ( indexes_tried[index] )
         {
         
            index = rand() % machine_size;  // get random index
         
         } // end while random number is a repeat
      
         saftey_counter++;
      
         if( saftey_counter == machine_size) // should never happen
         {
         
            cerr << "ERROR: Pick has exceeded machine size." << endl;
            break; // if so, infinite while - exit program
         
         } // end if saftey break
      
      } // end while ball is already picked
   
      indexes_tried[index] = TRIED;  // flag index as tried
   
      balls_picked++;  // increment the amount of balls picked
   
      return ( contents[index] );
   
   } // end LottoMachine::Pick ( void )
         
         
   unsigned char
   LottoMachine::Get_machine_size ( void )
   //------------------------------------------------------------------------
   // Returns value of machine size
   //------------------------------------------------------------------------
   {
      return( machine_size ); // simple accessor method returns machine_size
      
   } // end  LottoMachine::Get_machine_size ( void )


   int  
   LottoMachine::Get_balls_picked ( void )
   //------------------------------------------------------------------------
   // returns value of balls_picked
   //------------------------------------------------------------------------
   {
      return( balls_picked ); // simple accessor method returns balls_picked
      
   } // end LottoMachine::Get_balls_picked ( void )

   void 
   LottoMachine::Reset ( int index )
   //------------------------------------------------------------------------
   // resets the one ball of index in the Lotto Machine
   // does not reset balls_picked value to zero
   // use this reset when resetting for dupilcate picks
   //------------------------------------------------------------------------
   {
      if ( index < machine_size)   // if in bounds
      {        
         contents[index].Reset();  // reset ball
         indexes_tried[index] = 0; // reset indexes_tried
      } // if in bounds
      else
      {
         cerr  << "ERROR: passed out of bounds index to Reset ( int index )"
               << endl;

      } // end else out of bounds
   } // end LottoMachine::Reset ( int index )

   void  
   LottoMachine::Reset ( void )
   //------------------------------------------------------------------------
   // resets all balls to not picked and balls_picked value to zero 
   //------------------------------------------------------------------------
   {
      int i = 0;
      
      for (i = 0; i < machine_size; i++)   // reset contents
      {
         contents[i].Reset();
      } // end for reset
      
      for ( i = 0; i < machine_size; i++ )
      {
         indexes_tried[i] = 0;
      } // reset indexes_tried
      
      balls_picked = 0;
      
   } // end LottoMachine::Reset ( void )



#ifdef LMACH_TEST
   //------------------------------------------------------------------------
   // constant variables for unit test
   //------------------------------------------------------------------------
   const unsigned char LOTTO_SIZE = 49;

   const int NO_LOTTO_PICKS = 6;

   const unsigned char CASH3_SIZE = 10;

   const int NO_CASH3_PICKS = 3;

   int  
   main ( void )
   //------------------------------------------------------------------------
   // Unit test of the class LottoMachine
   // creates 2 LottoMachines of differrent sizes
   // picks the appropriate number of balls for each machine
   // shows the valus of the balls; resets and runs process again
   //------------------------------------------------------------------------
   {  
      int i = 0;
      cout << "Unit test of class LottoMachine\n\n";
      cout << "Create 2 LottoMachine objects of different sizes:\n\n";
      
      LottoMachine myLotto( LOTTO_SIZE );   // to test this size lotto
      LottoMachine myCash3( CASH3_SIZE );   // to test this size lotto

      Ball lotto_test_card[NO_LOTTO_PICKS];  // balls picked for lotto
      Ball cash3_test_card[NO_CASH3_PICKS];  // balls picked for cash3

      cout << "\n<<<<<<< Testing Lotto Machine: size Lotto >>>>>>>\n\n";
   
      while( myLotto.Get_balls_picked() < NO_LOTTO_PICKS )
      {
      
         lotto_test_card[i] = myLotto.Pick();
         i++;
      
      } // end while
   
      cout << "\tShow 1st picks:   ";
   
      for (i = 0; i < NO_LOTTO_PICKS; i++) // show balls
      {
         lotto_test_card[i].Show();
      } // end for show
   
      cout << "\n\nRESET Lotto Machine\n";  
   
      myLotto.Reset();
      i = 0;
   
      while( myLotto.Get_balls_picked() < NO_LOTTO_PICKS )
      {
      
         lotto_test_card[i] = myLotto.Pick();
         i++;
      
      } // end while
   
      cout << "\n\tShow 2nd picks:   ";
   
      for (i = 0; i < NO_LOTTO_PICKS; i++) // show balls
      {
         lotto_test_card[i].Show();
      } // end for show
   
      cout << "\n\n<<<<<<< Testing Lotto Machine: size Cash3 >>>>>>>\n\n";
   
      i = 0;
   
      while( myCash3.Get_balls_picked() < NO_CASH3_PICKS )
      {
      
         cash3_test_card[i] = myCash3.Pick();
         i++;
      
      } // end while
   
      cout << "\tShow 1st picks:   ";
   
      for (i = 0; i < NO_CASH3_PICKS; i++) // show balls
      {
         cash3_test_card[i].Show();
      } // end for show
   
      cout << "\n\nRESET Cash3 Machine\n";  
   
      myCash3.Reset(); 
      i = 0;
   
      while( myCash3.Get_balls_picked() < NO_CASH3_PICKS )
      {
      
         cash3_test_card[i] = myCash3.Pick();
         i++;
      
      } // end while
   
      cout << "\n\tShow 2nd picks:   ";
   
      for (i = 0; i < NO_CASH3_PICKS; i++) // show balls
      {
         cash3_test_card[i].Show();
      } // end for show
   
      cout << "\n\nEnd LottoMachine Unit Test\n" << endl;
   
      return EXIT_SUCCESS;
   
   }

#endif

// end lmach.cpp
