// ***************************************************************************
//       
// File:        playslip.cpp
// Author:      Bettina McCormick
//          
// Project:     Midterm: COP 22224 VCC; Summer 1999
//       
// Created:     June 18, 1999
// Updated:     June 23, 1999
// Date due:    June 28, 1999
//          
// Description: Implementation for the PlaySlip object that is used with  
//              the ball and Lotto machine objects in a simulation of a 
//              lotto machine.
//                
//              Added to midterm for enhancement of lotto main and to 
//              learn the creation of a class with no template.  Have 
//              found that this design will translate to an inheritance 
//              model - a learning goal for my next project.
//          
// ***************************************************************************
      
   // define PLAYSLIP_TEST in preprocessor definitions to run unit test      
            
   #include "playslip.h"
         
            
   PlaySlip::PlaySlip( unsigned char t, unsigned char p )
   //------------------------------------------------------------------------
   // initializes the object
   // calls Set_values to set private data
   //------------------------------------------------------------------------
   {
      Set_values( t, p ); // initializes to MAX_PICKS & MAX_PANELS 
      
      #ifdef PLAYSLIP_TEST // comment to output
      cout << "Constructor: created an instance of PlaySlip" << endl;
      #endif   
   } // end Constructor
         
         
   PlaySlip::~PlaySlip() 
   // cleans up the object
   {
      // no dynamic memory to deallocate
      #ifdef PLAYSLIP_TEST // comment to output
      cout << "Destructor: PlaySlip object dropped out of scope" << endl;
      #endif 
   
   } // end Destructor
         
         
   void 
   PlaySlip::Set_values ( unsigned char t, unsigned char p )
   //------------------------------------------------------------------------
   // sets type, panels & lm_size
   //------------------------------------------------------------------------
   {
      type = t;
      
      panels = p;
      
      switch ( type )      // determine size and no of picks
      {
         case NO_CASH3_PICKS:
            size = (unsigned char ) CASH3_SIZE;
            no_picks = ( unsigned char ) NO_CASH3_PICKS;
            break;
         
         case NO_PLAY4_PICKS:
            size = (unsigned char ) PLAY4_SIZE;
            no_picks = ( unsigned char ) NO_PLAY4_PICKS;
            break;
            
         case NO_FANTASY5_PICKS:
            size = (unsigned char ) FANTASY5_SIZE;
            no_picks = ( unsigned char ) NO_FANTASY5_PICKS;
            break;
            
         case NO_LOTTO_PICKS:
            size = (unsigned char ) LOTTO_SIZE;
            no_picks = ( unsigned char ) NO_LOTTO_PICKS;
            break;
            
         default: // shouldn't happen, but just in case :)
            cerr << "ERROR: slip type out of bounds" 
                 << "       PlaySlip::Set_values" << endl;
            break;
            
      } // end switch ( type )
         
   } // end Set_values ( unsigned char t, unsigned char p )
      
      
   char * 
   PlaySlip::Get_type ( void ) 
   //------------------------------------------------------------------------
   // returns a string describing type
   //------------------------------------------------------------------------
   {
      char * slip_name[4] = {"Cash3", "Play4", "Fantasy5", "Lotto"};
         
      return ( slip_name[type - 3] ); // returns the coresponding string
         
   } // end Get_type ( void ) 
         
      
   int 
   PlaySlip::Get_panels ( void )
   //------------------------------------------------------------------------
   // returns no of panels to be filled
   //------------------------------------------------------------------------
   {  
      return ( static_cast<int> ( panels ) ); // returns panels as an int
         
   } // end PlaySlip::Get_panels ( void )
         
         
   void  
   PlaySlip::Fill_slip ( void )
   //------------------------------------------------------------------------
   // fills the slip with lm picks
   //------------------------------------------------------------------------
   {
     int    ball_index = 0, // index to ball for when reset machine for dups
            p_idx = 0, // panel index
            b_idx = 0; // ball index
            
      LottoMachine lmPick_generator ( size ); // machine to generate picks
         
      if ( lmPick_generator.Get_machine_size() > PLAY4_SIZE )
      { // no duplicate numbers picked
         for (p_idx = 0; p_idx < panels; p_idx++)
         {
            b_idx = 0;
            while( lmPick_generator.Get_balls_picked() < no_picks )
            {
               // bubble in slip with balls picked     
               slip_contents[p_idx][b_idx] = lmPick_generator.Pick();
               b_idx++;
            } // end while 
            lmPick_generator.Reset();
         } // end for each panel
      } // end if Lotto or Fantasy5
      else
      { // duplicate numbers allowed in these picks
         for (p_idx = 0; p_idx < panels; p_idx++)
         {
            b_idx = 0;
            while( lmPick_generator.Get_balls_picked() < no_picks )
            {
               // bubble in slip with balls picked     
               slip_contents[p_idx][b_idx] = lmPick_generator.Pick();
               //special_count++;
               ball_index = static_cast<int> // index is ball value in this case
                           ( slip_contents[p_idx][b_idx].Get_number() ); 
               // reset this ball to all to be picked again
               lmPick_generator.Reset( ball_index );
               b_idx++;
            } // end while 
            lmPick_generator.Reset();
         } // end for each panel
      } // end else
      
   } // end Fill_slip ( void )
      
      
   void  
   PlaySlip::Show ( void )
   //------------------------------------------------------------------------
   // shows the Play Slip values to screen
   //------------------------------------------------------------------------
   {
      int   p_idx = 0,
            b_idx = 0; // for incrementation 
         
      unsigned char lotto_num = 0;
         
      if ( (type - 3) < 0 ) 
      { // prevent bad memory access
         cerr  << "ERROR: trying to access slip_name out of bounds\n" 
               << "       PlaySlip::Show" << endl;
      }
      else
      { // go for it
         if ( panels == 1 )
         {  // if singular tense
            cout  << "Your Lottery Numbers for this \"" 
                  << Get_type() << "\" Play Slip of " 
                  << Get_panels()
                  << " panel are: \n\n";
         } // end if singular tense
         else 
         {  // plural tense
            cout  << "Your Lottery Numbers for this \"" 
                  << Get_type() << "\" Play Slip of " 
                  << Get_panels()
                  << " panels are: \n\n";
         } // end else plural tense
      } // end else go for it
         
      for (p_idx = 0; p_idx < panels; p_idx++)
      { // for each panel
         cout << "\t    <<<  Panel " << (p_idx + 1) << "  >>>";
         for ( b_idx = 0; b_idx < no_picks; b_idx++ )
         { // for each pick
            slip_contents[p_idx][b_idx].Show();
               
         } // end for each pick
            
         cout << "\n" << endl;
            
      } // end for each panel
         
   } // end Show ( void )
      
      
   void  
   PlaySlip::Show ( ofstream & outfile ) 
   //------------------------------------------------------------------------
   // takes a pointer to an open file 
   // shows the Play Slip values to a file
   // example of overloading the method Show
   //------------------------------------------------------------------------
   {
      static int FILE_HEADER = 0;  // flag to write header to file one time    
      
      int   p_idx = 0,
            b_idx = 0; // for incrementation 
      
      unsigned char lotto_num = 0;
      
      FILE_HEADER++;  // increment each time this method called
      
      if ( (type - 3) < 0 ) 
      { // prevent bad memory access
         cerr  << "ERROR: trying to access slip_name out of bounds\n" 
               << "       PlaySlip::Show( ofstream & outfile )" << endl;
      }
      if ( !outfile.is_open() )
      { // if file passed is not open 
         cerr  << "ERROR: file passed is not open\n" 
               << "PlaySlip::Show( ofstream & outfile )" << endl;
      }
      else
      { // go for it
         if ( FILE_HEADER == 1 ) // first time this file is being written
         {
            outfile  << "\tYour \"Quick Pick\" Numbers...     Good luck!!\n\n" 
                     << endl;
         }
         
         if ( panels == 1 )
         {  // if singular tense
         
            outfile << "Your Lottery Numbers for this \"" 
               << Get_type() << "\" Play Slip of " 
               << Get_panels()
               << " panel are: \n\n";
         }
         else 
         {  // plural tense
         
            outfile << "Your Lottery Numbers for this \""
               << Get_type() << "\" Play Slip of " 
               << Get_panels()
               << " panels are: \n\n";
         }
      
      } // end else go for it
      
      for (p_idx = 0; p_idx < panels; p_idx++)
      { // for each panel
         outfile << "\t    <<<  Panel " << (p_idx + 1) << "  >>>";
            
         for ( b_idx = 0; b_idx < no_picks; b_idx++ )
         { // for each pick
            lotto_num = slip_contents[p_idx][b_idx].Get_number();
            outfile << setw( 5 ) << static_cast<int> (lotto_num);
               
         } // end for each pick
            
         outfile << "\n\n" << endl;
            
      } // end for each panel
         
   } // end PlaySlip::Show ( ofstream & outfile ) 
         
         
   void  
   PlaySlip::Reset ( void )
   //------------------------------------------------------------------------
   // resets the Play Slip to largest lotto - same as default of constructor
   //------------------------------------------------------------------------
   {
      Set_values ( MAX_PICKS, MAX_PANELS );// resets to MAX_PICKS, MAX_PANELS 
   
   } // end PlaySlip::Reset ( void ) 
         
         
         
#ifdef PLAYSLIP_TEST
   
   int   
   main ( void )
   //------------------------------------------------------------------------
   // unit test of all the methods of playslip
   //------------------------------------------------------------------------
   {  
      bool  usefile = false; // assume false
      int   i = 0;  // for incrementation
      int   my_type[] = { 3, 4, 5, 6 };  // types of play slips
      unsigned char my_panels = 5;  // num of panels
      PlaySlip my_playslip( (unsigned char) my_type[3], my_panels );
      
      // constructor opens file to append output to
      ofstream qckpckDataFile( "quickpick.dat", ios::out ); 
      
      if ( !qckpckDataFile )
      { // if file fails note it and leave flag false
         cerr << "Output file will not be used for this execution." << endl;
      }
      else
      { // else make flag true
         usefile = true;
      }
      
      cout << "Unit test of class PlaySlip\n\n";
      
      for (i = 0; i < 4; i++)
      { // to get wide rage of possible scenarios
         my_playslip.Set_values( my_type[i], my_panels );
         my_playslip.Fill_slip();
         my_playslip.Show();
         
         if ( usefile )
         {  // if file suceeded, use file
            cout << "Writing data to file \n"<< endl;
            my_playslip.Show( qckpckDataFile );
         }
         
         my_playslip.Reset();
      } // end for 
      
        
      if ( usefile )
      {
         cout  << "The data for this execution has also been written to file" 
               << " \"quickpick.dat\"" << endl;
      }
         
      cout << "\nEnd PlaySlip Unit Test\n" << endl;
      // ofstream deconstructor closes the file
      return  EXIT_SUCCESS;  
      
   } // end main
      
#endif
      
// end playslip.cpp
