/* title  hough.c
/*****************************************************************************\

/*------------------------------- A U T H O R  -------------------------------*\
	Bettina McCormick
	CAP 4453 - Robot Vision - Fall 1998
	Assignment 2B HoughTransformer  - Due Date: October 5, 1998
	Version 2.8  - last updated  October 2, 1998
\*----------------------------------------------------------------------------*/

/*------------------------ D E S C R I P T I O N  ----------------------------*\
	This program implements the Hough Transformer for a specific input.  It fits
	the edges in this input to a linear parametric equation.  The first output is
	the canny image that was read in as input. The second output is the strongest
	line found in the cannyimage.  The third output is both the second and third
	strongest lines in the canny image.
\*----------------------------------------------------------------------------*/

/*--------------------- R E G A R D I N G _ I N P U T  -----------------------*\
  Input for this program is the output from assignment 2A - the Canny Edge
  detection of the Half gallon image (halfgal34.raw).  The
  dimensions of this image are 256 by 256.  The user enters the input on the
  command line followed by the three output files, the integer number for the
  divisions of Theta and the integer number for the divisions of D. (7 comand
  line arguments in that order)
\*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h> /* not recognized by UNIX - use the -lm commmand instead */

/* #define  DEBUG */
#define  X_DIMENSION	  		256    	/* x dim of image */
#define  Y_DIMENSION	  		256    	/* y dim of image */
#define  D_DIMENSION       400      /* D dim of votes array */
#define  T_DIMENSION       136      /* theta dim of votes array */
#define  V_DIMENSION       260      /* votes dim of who_voted array */
#define  NEIGHBOR_CHECK    1
#define  EDGE			      255      /* for easier readability */
#define  T_upper_bound 		0.785398163      /* range of theta -3pi/4 to pi/4 */
#define  T_lower_bound 		-2.35619449
#define  Pi			      	3.1415927      /* value for pi */
#define  T_max 				3.1415927      /* theta maximum range */
#define  ZERO			      0       /* for easier readability */


   int main (int argc, char ** argv)
   {
   /*----------------  V A R I A B L E  D E C L A R A T I O N S  ----------------*/
   
      FILE 	*infp1,                   /* pointer to input file */
         *outfp1,                  /* pointer to output file */
         *outfp2,                  /* pointer to output file */
         *outfp3,                  /* pointer to output file */
         *outfp_mytemp,            /* pointer to my temp file */
         *fopen();                 /* pointer to function */
   
      struct xy {
         char 	x_coord,
         y_coord;
      };
   
      typedef  struct xy xy_votes;
   
      int 	Ay, Bx, Iy, Jx, X, Y, Z,    /* for incrementation */
      Ti,           /* theta index */
      Di,           /* D index */
      winning_D,
      winning_T,
      votes_max = 0,         /* votes maximum */
      votes_max_sum = 0,
      votes_temp = 0,
      X_max = 0,
      Y_max = 0,
      X_max2 = 0,
      Y_max2 = 0,
      divs_d,   /* divisions of D for D dimension -read in on command line */
      divs_t;   /* divisions of T for T dimension -read in on command line */
   
   
      int   input_image[Y_DIMENSION][X_DIMENSION];
   				/* 2D array to hold input image */
   
      int   first_place_image[Y_DIMENSION][X_DIMENSION];
   				/* 2D array to hold first place winner image */
   
      int   sec_thr_image[Y_DIMENSION][X_DIMENSION];
   				/* 2D array to hold 2nd & 3rd place winner images */
   
      int   votes[D_DIMENSION][T_DIMENSION];
   				/* 2D array to hold input image */
   
      xy_votes  who_voted[D_DIMENSION][T_DIMENSION][V_DIMENSION];
   					/* 2D array to hold x & y data */
   
      double  	T,             /* theta */
      T_step = 0,        /* theta step */
      T_temp1 = 0,	       	/* theta temp */
      T_temp2 = 0,	       	/* theta temp */
      D,             	/* D */
      D_max = 0,         /* D maximum */
      D_min = 0,         /* D minimum */
      D_offset = 0,      /* D offset - abs value of D_min */
      D_step = 0;        /* D step */
   
   
   /*--------------------------  B E G I N  M A I N  ----------------------------*/
   
      if(argc < 7)   /* test arguments */
      {
         fprintf(stderr, "\nYou must enter 6 command-line arguments following ");
         fprintf(stderr, "the executable:\n\n");
         fprintf(stderr,
                "C:\\>executable input output1 output2 output2 Theta_divisions D_divisions\n");
         exit(1);
      };
   
   /*-----------------------------------------------------------------open files */
      if((infp1 = fopen(argv[1], "r")) == NULL)   /* opens 1st input file */
      {
         fprintf(stderr, "Error opening input file.");
         exit(1);
      };
   
      if((outfp1 = fopen(argv[2], "w")) == NULL)   /* opens output file */
      {
         fprintf(stderr, "Error opening first output file.");
         exit(1);
      };
      if((outfp2 = fopen(argv[3], "w")) == NULL)   /* opens output file */
      {
         fprintf(stderr, "Error opening second output file.");
         exit(1);
      };
   
      if((outfp3 = fopen(argv[4], "w")) == NULL)   /* opens output file */
      {
         fprintf(stderr, "Error opening third output file.");
         exit(1);
      };
   
   #ifdef DEBUG
   if((outfp_mytemp = fopen("temp1.dat", "w")) == NULL)   /* opens my temp file */
   {
   	fprintf(stderr, "Error opening my temp file.");
   	exit(1);
   };
   #endif
   /*---------------------------------assign command line arguments to variables */
      divs_t = atoi(argv[5]);        /* SIGMA read from command line */
      if (divs_t > T_DIMENSION)
      {
         printf("You entered a value greater than %d.", T_DIMENSION);
         printf("Program halted - run again with a lesser T division.");
         exit(1);
      }
      divs_d = atoi(argv[6]);        /* SIGMA read from command line */
      if (divs_d > D_DIMENSION)
      {
         printf("You entered a value greater than %d.", D_DIMENSION);
         printf("Program halted - run again with a lesser D division.");
         exit(1);
      }
   
   
   /*--------------------------------------------------------read in canny image */
      for (Ay = 0; Ay < Y_DIMENSION; Ay++)
      {
         for (Bx = 0; Bx < X_DIMENSION; Bx++)
         {
            input_image[Ay][Bx]  =  getc (infp1);
            input_image[Ay][Bx]  &= 0377;
         } /* end for Bx */
      } /* end for Ay */
   
   /*---------------------------------------------------------output canny image */
      for (Ay = 0; Ay < Y_DIMENSION; Ay++)
      {
         for (Bx = 0; Bx < X_DIMENSION; Bx++)
         {
            fprintf(outfp1,"%c",(char)input_image[Ay][Bx]);
         } /* end for Bx */
      } /* end for Ay */
   
   /***********************     BEGIN PROCESSING IMAGE    ************************/
      T_step = (T_max / divs_t);
   /*-----------------------------------------------------------scan canny image */
   /*----------------------------------------------------------collect maximum D */
      for (Ay = 0; Ay < Y_DIMENSION; Ay++)     /* range of theta -3pi/4 to pi/4 */
      {
         for (Bx = 0; Bx < X_DIMENSION; Bx++)
         {
            if (input_image[Ay][Bx] == EDGE)
            {
               T = T_lower_bound;
            
               while (T < T_upper_bound)
               {
                  D = ((Ay * sin(T)) + (Bx * cos(T)));
                  if (D > D_max)
                  {
                     D_max = D;
                     Y_max = Ay;
                     X_max = Bx;
                     T_temp1 = T;
                  }
                  if (D < D_min)
                  {
                     D_min = D;
                     Y_max2 = Ay;
                     X_max2 = Bx;
                     T_temp2 = T;
                  }
                  T = T + T_step;
               }
            }
         } /* end for Bx */
      } /* end for Ay */
   /*------------------------------------------------print D_max to my temp file */
   #ifdef DEBUG
   fprintf(outfp_mytemp,"1st D_max = %7.3f\tT = %7.3f [%d][%d]\n",
   	D_max, T_temp1, Y_max, X_max);
   fprintf(outfp_mytemp,"1st D_min = %7.3f\tT = %7.3f [%d][%d]\n",
   	D_min, T_temp2, Y_max2, X_max2);
   #endif
   /*--------------------------------------------------------------collect votes */
      D_offset = fabs(D_min);
      D_max = D_max + D_offset;
      D_step = (D_max / divs_d);
   
   
      for (Ay = 0; Ay < Y_DIMENSION; Ay++)
      {
         for (Bx = 0; Bx < X_DIMENSION; Bx++)
         {
            if (input_image[Ay][Bx] == EDGE)
            {
               if ((Ay == 12) && (Bx == 68))
                  fprintf(outfp_mytemp,"12,68[%d][%d]\n",
                         Ay, Bx);
               T = T_lower_bound;
               Ti = ZERO;
               Di = ZERO;
               while (T < T_upper_bound)
               {
                  D = ((Ay * sin(T)) + (Bx * cos(T)));
                  D = D + D_offset;
                  Di = (int) (D / D_step);
                  votes[Di][Ti]++;
                  who_voted[Di][Ti][votes[Di][Ti]].x_coord = Bx;
                  who_voted[Di][Ti][votes[Di][Ti]].y_coord = Ay;
                  T = T + T_step;
                  Ti++;
               }
            }
         } /* end for Bx */
      } /* end for Ay */
   
   /*--------------------------------------------collect maximum votes 1st place */
      votes_max = 0;
      for (Ay = 0; Ay < D_DIMENSION; Ay++)
      {
         for (Bx = 0; Bx < T_DIMENSION; Bx++)
         {
            if (votes[Ay][Bx] > votes_max)
            {
               votes_max = votes[Ay][Bx];
               winning_D = Ay;
               winning_T = Bx;
            }
         } /* end for Bx */
      } /* end for Ay */
   
   #ifdef DEBUG
   fprintf(outfp_mytemp,"1st votes_max = %7.3d\nwin_D = %d\nwin_T = %d\n",
   	votes_max, winning_D, winning_T);
   #endif
   
   /*----------------------------------------------get x & y votes out 1st place */
      for (Z = 1; Z <= votes_max; Z++)
      {
         Y = (int)who_voted[winning_D][winning_T][Z].y_coord;
         X = (int)who_voted[winning_D][winning_T][Z].x_coord;
         first_place_image[Y][X] = EDGE;
      } /* end for Z */
      first_place_image[68][12] = 0;
   
   /*---------------------------------------------------zero out first votes max */
   
      for (Iy = -NEIGHBOR_CHECK; Iy <= NEIGHBOR_CHECK; Iy++)
      {
         for(Jx = -NEIGHBOR_CHECK; Jx <= NEIGHBOR_CHECK; Jx++)
         {
         
            votes[winning_D + Iy][winning_T + Jx] = ZERO;
         
         }
      }
   
      votes[winning_D][winning_T] = ZERO;
   
   /*--------------------------------------------collect maximum votes 2nd place */
      votes_max = 0;
      winning_D = 0;
      winning_T = 0;
   
      for (Ay = 0; Ay < D_DIMENSION; Ay++)
      {
         for (Bx = 0; Bx < T_DIMENSION; Bx++)
         {
            if (votes[Ay][Bx] > votes_max)
            {
               votes_max = votes[Ay][Bx];
               winning_D = Ay;
               winning_T = Bx;
            }
         } /* end for Bx */
      } /* end for Ay */
   
   #ifdef DEBUG
   fprintf(outfp_mytemp,"2nd place votes_max = %7.3d\nwin_D = %d\nwin_T = %d\n",
   	votes_max, winning_D, winning_T);
   #endif
   
   /*----------------------------------------------get x & y votes out 2nd place */
      for (Z = 1; Z <= votes_max; Z++)
      {
         Y = (int)who_voted[winning_D][winning_T][Z].y_coord;
         X = (int)who_voted[winning_D][winning_T][Z].x_coord;
         sec_thr_image[Y][X] = EDGE;
      } /* end for Z */
   /*-----------------------------------------------------zero out 2nd votes max */
      for (Iy = -NEIGHBOR_CHECK; Iy <= NEIGHBOR_CHECK; Iy++)
      {
         for(Jx = -NEIGHBOR_CHECK; Jx <= NEIGHBOR_CHECK; Jx++)
         {
         
            votes[winning_D + Iy][winning_T + Jx] = ZERO;
         
         }
      }
   
   /*--------------------------------------------collect maximum votes 2nd place */
      votes_max = 0;
      winning_D = 0;
      winning_T = 0;
   
      for (Ay = 0; Ay < D_DIMENSION; Ay++)
      {
         for (Bx = 0; Bx < T_DIMENSION; Bx++)
         {
            if (votes[Ay][Bx] > votes_max)
            {
               votes_max = votes[Ay][Bx];
               winning_D = Ay;
               winning_T = Bx;
            }
         } /* end for Bx */
      } /* end for Ay */
   
   #ifdef DEBUG
   fprintf(outfp_mytemp,"3rd place votes_max = %7.3d\nwin_D = %d\nwin_T = %d\n",
   	votes_max, winning_D, winning_T);
   #endif
   
   /*----------------------------------------------get x & y votes out 3rd place */
      for (Z = 1; Z <= votes_max; Z++)
      {
         Y = (int)who_voted[winning_D][winning_T][Z].y_coord;
         X = (int)who_voted[winning_D][winning_T][Z].x_coord;
         if (sec_thr_image[Y][X] == EDGE)
         {
            fprintf(outfp_mytemp,"alert - 3rd overwriting 2nd\n" );
         }
         sec_thr_image[Y][X] = (EDGE - 100);
      
      } /* end for Z */
   
   /*-----------------------------------------collect maximum votes of neighbors */
      for (Ay = NEIGHBOR_CHECK; Ay < (D_DIMENSION - NEIGHBOR_CHECK); Ay++)
      {
         for (Bx = NEIGHBOR_CHECK; Bx < (T_DIMENSION - NEIGHBOR_CHECK); Bx++)
         {
            votes_temp = 0;
            for (Iy = -NEIGHBOR_CHECK; Iy <= NEIGHBOR_CHECK; Iy++)
            {
               for(Jx = -NEIGHBOR_CHECK; Jx <= NEIGHBOR_CHECK; Jx++)
               {
                  votes_temp = votes_temp + votes[Ay +Iy][Bx + Jx];
               }
            }
            if (votes_temp > votes_max_sum)
            {
               votes_max_sum = votes_temp;
               winning_D = Ay;
               winning_T = Bx;
            }
         } /* end for Bx */
      } /* end for Ay */
   
   #ifdef DEBUG
   fprintf(outfp_mytemp,"scanned votes_max_sum = %7.3d\nwin_D = %d\nwin_T = %d\n",
   votes_max_sum, winning_D, winning_T);
   #endif
   
   /*************************     END PROCESSING IMAGE    ************************/
   /*----------------------------------------------output 1st place winner image */
   
      for (Ay = 0; Ay < Y_DIMENSION; Ay++)
      {
         for (Bx = 0; Bx < X_DIMENSION; Bx++)
         {
            fprintf(outfp2,"%c",(char)first_place_image[Ay][Bx]);
         } /* end for Bx */
      } /* end for Ay */
   
   /*---------------------------------------output 2nd & 3rd place winners image */
   
      for (Ay = 0; Ay < Y_DIMENSION; Ay++)
      {
         for (Bx = 0; Bx < X_DIMENSION; Bx++)
         {
            fprintf(outfp3,"%c",(char)sec_thr_image[Ay][Bx]);
         } /* end for Bx */
      } /* end for Ay */
   
   /*--------------------------------------------------------------  close files */
      fclose(infp1);
      fclose(outfp1);
      fclose(outfp2);
      fclose(outfp3);
   
      return EXIT_SUCCESS;
   } /* end main a2b.c */

