/* ************************************************************************** */
/*      			Bettina McCormick, January 1998, CSIII, Dr. Leeson	      */
/*                                                                            */
/*       Huffman Code Compression Project  - Due Date: February 9, 1998       */
/*                                                                            */
/*           encode.c - Version 9.1 - last updated January 30, 1998           */
/*                                                                            */
/*                                                                            */
/*       This program takes a text file:  creates a frequecy table,           */
/*      a huffman tree using a heap for the implementation, encodes           */
/*      the file to a intermediate file, and outputs two final files -        */
/*      .tre and .cod: the .tre is the frequency table information and        */
/*      the .cod is the coded file converted to bits for compression.         */
/*                                                                            */
/* ************************************************************************** */

#include "heap.h"

   int main( int argc, char **argv )
   {
      int  filedata,		                  			/* to store data from file */
      charnum,
      filenum = 0,     /* to store size of original ASCII characters in chars */
      j;                                                /* for incrementation */
   
      unsigned int freqtable[MAXARRAY] = {0};        	 /* to hold frequencies */
   
      unsigned long int k,
      l = 0,
      codelength = 0;/*to store number of char in coded .mid file*/
   
      char codebits,/* to store single char of bits matchin 8 char of coded file */
      str1[MAXSTRING] = {'\0'},    /* for filename conversion with extensions */
      str2[MAXSTRING] = {'\0'},    /* for filename conversion with extensions */
      str3[MAXSTRING] = {'\0'},    /* for filename conversion with extensions */
      tre[] = ".tre",              /* for filename conversion with extensions */
      mid[] = ".mid",              /* for filename conversion with extensions */
      cod[] = ".cod",              /* for filename conversion with extensions */
      nullstring[MAXCODE] = {'\0'},
      codeary[MAXARRAY] = {'\0'};    /* dummy array to simulate pass by value */
   
      FILE  *in_fp, *out_fp, *tmp_fp;                             /* pointers to files */
   
      node node1, *ptrnode, *ptrnode1, *ptrnode2;/* to hold file data and treelinks*/
   
      heap heap1, *pheap;        /* to hold the nodes in a heap implementaion */
   
      code huffcode[MAXARRAY];          /* array of code structures (strings) */
   
   
   /***********************       begin main          *************************/
   
      if(argc < 3)   /* test arguments */
      {
         fprintf(stderr, "\nYou must enter 2 command-line arguments following ");
         fprintf(stderr, "the executable:\n\n");
         fprintf(stderr,
                "C:\\>executable  input_file  output_files_prefix \n");
         exit(1);
      };
   
      strcpy(str1, argv[2]); strcpy(str3, argv[2]);
   							  /* copies prefix to temp variable for concatonation */
   
      strcat(str1, tre);  strcat(str3, cod);
   											/* adds suffix to prefix for output files */
   
      if( (in_fp = fopen(argv[1], "r")) == NULL)      /* opens input.txt file */
      {
         fprintf(stderr, "Error opening file.");
         exit(1);
      }
      if ( (out_fp = fopen(str1, "w")) == NULL)      /* opens output.tre file */
      {
         fprintf(stderr, "Error opening file.");
         exit(1);
      }
   
   /****************************   make freqtable  ****************************/
   
      while ((filedata = getc(in_fp)) != EOF)
      {                          /* reads single char from file, saves as int */
         freqtable[filedata]++;                   /* increments contents by 1 */
         filenum++;
      }; /* end while */
   
   /********************** output freqtable to  file argv[2].tre **************/
   
      for (j = 0; j < MAXARRAY; j++)
      {
         fprintf(out_fp, "%d ", freqtable[j]);
      #ifdef DEBUG
      if (freqtable[j] != 0)
      printf("%c  %d\n", j, freqtable[j]);
      #endif
      }/* end for */
      printf("help");
      fprintf(out_fp, "%d", filenum); 
      fclose(out_fp);    		                                 /* closes file */
   
   /*****************************   make heap  ********************************/
   
      HeapInit(&heap1);                                   /* initialize heap1 */
      for (j = 0; j < (MAXARRAY); j++)
      {                      /* creates a node and puts pointer to it in heap */
         if (freqtable[j] !=0)
         {
            ptrnode = (node*) malloc(sizeof(node));
            ptrnode -> fvalue = freqtable[j];
            ptrnode -> cdata = (char)(j);
            ptrnode -> scode[0] = '\0';
            ptrnode -> leftc = NULL;
            ptrnode -> rightc = NULL;
            HeapInsert (&heap1, ptrnode, heap1.hcount+1);
            heap1.hcount++;
         }  /* end if */
      }   /* end for */
   
   #ifdef DEBUG
   HeapDisplay(&heap1, heap1.hcount);
   #endif
   
   /*************************    make Huffman Tree   **************************/
   
      while (heap1.hcount >1) /* removes nodes from heap and builds hufftree */
      {
         ptrnode1 = DeleteRoot(&heap1);
         heap1.hcount--;
         ptrnode2 = DeleteRoot(&heap1);
         heap1.hcount--;
         ptrnode = (node*) malloc(sizeof(node));
         ptrnode -> fvalue = (ptrnode1 -> fvalue + ptrnode2 -> fvalue);
         ptrnode -> cdata = '*';
         ptrnode -> scode[0] = '\0';
         ptrnode -> leftc = ptrnode1;
         ptrnode -> leftc -> scode[0] = '0';
         ptrnode -> leftc -> scode[1] = '\0';
         ptrnode -> rightc = ptrnode2;
         ptrnode -> rightc -> scode[0] = '1';
         ptrnode -> rightc -> scode[1] = '\0';
         HeapInsert (&heap1, ptrnode, heap1.hcount+1);
         heap1.hcount++;
      } /*end while */
   
   #ifdef DEBUG
   printf ("%s","\nthis be the tree:\n");
   PreorderTraversal(ptrnode);
   #endif
   
   /***********************   output Hufman code to an array    ***************/
   
      strcpy(huffcode[MAXARRAY].codestring, nullstring);/*initialize HuffCode array*/
      RetrieveCode(ptrnode, codeary, huffcode);/* output code to HuffCode array */
   
   /***** read and encode original file to argv[2].mid intermediate file ******/
   
      rewind(in_fp);                              /* resets file to beginning */
   
      tmp_fp = tmpfile();
   
      while ((filedata = getc(in_fp)) != EOF)
      {                          /* writes coded file in char to temp file */
         fputs(huffcode[filedata].codestring, tmp_fp);
         codelength += strlen(huffcode[filedata].codestring);
      #ifdef DEBUG
      printf("%s", huffcode[filedata].codestring);
      #endif
      }; /* end while */
   
   #ifdef DEBUG
   printf("\n%d\n", codelength);
   #endif
   
      fclose(in_fp);                                     /* closes file */
   
   /** generated the final binary argv[2].cod file from the intermediate file */
   
      rewind(tmp_fp);
      if ( (out_fp = fopen(str3, "wb")) == NULL)/* opens output .cod binary file */
      {
         fprintf(stderr, "Error opening file.");
         exit(1);
      };
   
      charnum = 0;
      codebits = 1;
      codebits <<= BYTE1;	        /* sets codebits bit value to '00000000' */
   
      while (l < codelength)
      {
         if ((codelength - l) < BYTE1)
         {
            k = (codelength - l);
            for (j = 0; j < k; j++)
            {
               filedata = fgetc(tmp_fp);
               l++;
               if (filedata == '0')
               {
                  codebits <<= 1;
               } /* end if */
               else
               {
                  codebits <<= 1;
                  codebits++;
               } /* end else */
            } /* end for */
            codebits <<= (BYTE1 - k);      /* shifts bits the rest of the way*/
            fputc(codebits, out_fp);
            charnum++;
         #ifdef DEBUG
         bit_print(codebits);
         #endif
         } /* end if */
         else
         {
            for (j = 0; j < BYTE1; j++)
            {
               filedata = fgetc(tmp_fp);
               l++;
               if (filedata == '0')
               {
                  codebits <<= 1;
               } /* end if */
               else
               {
                  codebits <<= 1;
                  codebits++;
               } /* end else */
            } /* end for */
            fputc(codebits, out_fp);
            charnum++;	  
         #ifdef DEBUG
         bit_print(codebits);
         #endif
            codebits = 1;
            codebits <<= BYTE1;
         } /* end else */
      } /* end while */                            /*  whew! that was fun!  */
   
      fclose(out_fp);    				                  /* closes file */
   
      printf("\n\n\tThis program has successfully generated two files: \n");
      printf("\n\t- the \"%s\" file contains the frequency table information; \n",
            str1);
      printf("\t- the \"%s\" file is the encoded version of \"%s\".\n", str3,
            argv[1]);
      printf("\n\n\tNow you can call the \"decode\" program!!\n\n");
   
      printf("\tC:\\>decode %s   %s   %s.out\n\n", str1, str3, argv[2]);
   
      return EXIT_SUCCESS;
   }
/***************************  end encode program   ****************************/


