/* ************************************************************************** */
/*      Bettina McCormick, January 1998, CSIII, Dr. Leeson, Assignment #1     */
/*                                                                            */
/*          heap.c - Version 2.5 - last updated January 30, 1998              */
/*                                                                            */
/*              Header file companion for Huffman Heap project -              */
/*             contains function code as described in heap.h file             */
/*                                                                            */
/* ************************************************************************** */

#include "heap.h"


   void HeapInit (heap *H)
   /* pointer to heap passed - precondition heap not yet used;
   postcondition: heap count set to zero */
   {
      H -> hcount = 0;
   }
   int HeapCount(heap *H)
   /* pointer to heap passed - returns size of heap */
   {
      return H -> hcount;
   }   /* end HeapInit */

   void HeapInsert (heap *H, node *ptr, int newindex)
   /* pointer to heap, pointer to node, and position in array passed -
   inserts new node in heap */
   {
      H -> heapary[newindex] = ptr;
      SettleUp(H, newindex);
   }  /* end HeapInsert */

   void HeapDisplay (heap *H, int index)
   /* displays heap for me */
   {
      int h;
      printf ("this be the heap:\n");
      for (h=1; h <= index; h++)
      {
         printf("%c  ", H -> heapary[h]->cdata);
         printf("%d\n", H -> heapary[h]->fvalue);
      }
   }

   void PreorderTraversal (node *N)
   /* displays tree for me */
   {
      if (N != NULL)
      {
         printf("%c  ", N -> cdata);
         printf("%d\n", N -> fvalue);
         PreorderTraversal (N->leftc);
         PreorderTraversal (N->rightc);
      }
   }

   void Swap(heap *H, int child, int parent)
   /* pointer to heap, two positions in array passed -
   postcondition: two nodes in positions passed are swapped in heap */
   {
      node *temp;
      temp = H ->heapary[child];
      H -> heapary[child]= H -> heapary[parent];
      H -> heapary[parent]= temp;
   }  /* end Swap */

   void SettleUp(heap *H, int newindex)
   /* pointer to heap, position in array passed - calls swap & itself;
   postcodition: heap is settled in relationship order - every node is
   "more powereful then its children and "less powerful" than its parent
   (smallest freq value is "powerful" in this heap) */
   {
      if (newindex > 1 &&
         (H -> heapary[newindex] -> fvalue < H -> heapary[newindex/2] -> fvalue))
      {
         Swap(H, newindex, newindex/2);
         SettleUp(H, newindex/2);
      }
   }   /* end SettleUp */

   node *DeleteRoot(heap *H)
   /* pointer to heap passed - calls swap & settledown;
   precondition: heap as more than one node;
   postcondition: root node is removed from heap and pointer to it is returned;
   heap is settled in relationship order  */
   {
      node *temp;
      Swap(H, 1, H -> hcount);
      SettleDown(H, 1, H -> hcount-1);
      temp = H -> heapary[H -> hcount];
      H -> heapary[H -> hcount] = NULL;
      return temp;
   }   /* end Deleteroot */

   void SettleDown(heap *H, int top, int bottom)
   /* pointer to heap, two positions in array passed - calls swap & itself;
   postcodition: heap is settled in relationship order - every node is
   "more powereful then its children and "less powerful" than its parent
   (smallest freq value is "powerful" in this heap) */
   {
      int child = 2*top;
      if  (child < bottom && H -> heapary[child + 1] -> fvalue <
          H -> heapary[child] -> fvalue)
      {
         ++child;
      }
      if  (child <= bottom && H -> heapary[top] -> fvalue >
          H -> heapary[child] -> fvalue)
      {
         Swap (H, top, child);
         SettleDown (H, child, bottom);
      }
   }  /* end SettleDown */

   void RetrieveCode(node *ptr, char C1[MAXCODE], code CD1[MAXARRAY])
   /* pointer to node, array of char & array of strings passed;
   tree is traversed and code is concatonated ans stored to array of
   strings corresponding with ascii value */
   {
   /*  #ifdef DEBUG
   FILE *code_fp;
   #endif */
      char c[MAXCODE];
      strcpy (c, C1);
   /*  #ifdef DEBUG
   if( (code_fp = fopen("code.tre", "r")) == NULL)
   {
   fprintf(stderr, "Error opening file.");
   exit(1);
   };
   #endif	 */
      if (ptr -> leftc != NULL && ptr ->rightc != NULL)
      {
         strcat (c, ptr -> scode);
         RetrieveCode(ptr -> leftc, c, CD1);
         RetrieveCode(ptr -> rightc, c, CD1);
      }
      else /* if node has no children - then write to output file */
      {
         strcat (c, ptr -> scode);
         strcpy(CD1[ptr -> cdata].codestring, c);
      }
   }   /* end RetrieveCode */

   void bit_print (unsigned char a)
   /* displays bits for me */
   {
      int i;
      int n = sizeof(char) * BYTE1;
      char mask = 1 << (n - 1);
      for (i = 1; i <= n; ++i)
      {
         putchar(((a & mask) == 0) ? '0' : '1');
         a <<=1;
         if (i % BYTE1 == 0 && i < n)
            putchar(' ');
      }
   }

   void DecodeNode(unsigned long int *L, node *P, FILE *infp, FILE *outfp)
   /* traverses the codetree recursively the leaf of the code to the out file */
   {
      char C;
      if ((P -> leftc == NULL) && (P -> rightc == NULL))
      {
         fprintf(outfp, "%c", P -> cdata);
      } /* end if */
      else
      {
         C = fgetc(infp);
         ++*L;
         if (C == '0')
         {
            DecodeNode(L, P -> leftc, infp, outfp);
         } /* end if */
         else if (C == '1')
         {
            DecodeNode(L, P -> rightc, infp, outfp);
         } /* end else */
      
      } /* end else */
   } /* end DeleteNode */
/*************************************end heap.c*******************************/

