/* ************************************************************************** *\
                           Bettina McCormick
          April 1998, COT 4210 - Discrete Structures, Extra Credit
                       Due Date: April 13, 1998
					 Version 7.3 - last updated  April 7, 1998
\* ************************************************************************** */

/*------------------------ D E S C R I P T I O N  ----------------------------*\
    This program implements the Powerset Construction for converting an
 arbitrary NFA to its equivalent DFA in reduced form (only reachable states).
 The input/output specifications are explained in detail in the file,
 "Prog-1-S98.doc".
\*----------------------------------------------------------------------------*/

/*--------------------- R E G A R D I N G _ I N P U T  -----------------------*\
    This program will take as many machines as you would like to put in the
    input file.  However, for analysis of the complimenting of a machine -
    the upper bound is 10 machines per input file.
\*----------------------------------------------------------------------------*/

/*--------- N O T E:  to whom it may concern and to my future self -----------*\
 The coding of this program is the sole accomplishment of Bettina McCormick.
 The methods used may be very "brute force"; hopefully, time and practice will
 reaveal a more clever method(s) for accomplishing this program.  Until then I
 have given my best effort to make this code understandable to my future self...
 and any who may venture to read it.
\*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/*-------------------------- # D E F I N E S ---------------------------------*/
/*#define DEBUG*/                    /* for my debugging */
#define BLANK         " "            /* for easier readability */
#define TRUE            1            /* for easier readability */
#define FALSE           0            /* for easier readability */
#define CAP_PHI         0            /* for easier readability */

/*----------------------------------------------------------- for Menu Choice */
#define COMPLIMENT                1      /* for easier readability */
#define REVERSE                   2      /* for easier readability */
#define COMPLIMENT_OF_REVERSE     3      /* for easier readability */
#define REVERSE_OF_COMPLIMENT     4      /* for easier readability */
#define QUIT                      5      /* for easier readability */

/*------------------------------------------------------------ maximum bounds */
#define INP_LENGTH    200            /* maximum input line length */
#define SIGMA_STAR    128            /* maximum alphabet in input */
#define MAX_STATES     32            /* maximum states in input */
   /* used (MAX_STATES+1) to simulate 1-based array for ease of readability*/

/*----------------------------------------  column indexes for machine arrays */
#define STATE_USED    130            /* column to indicate if state is in NFA */
#define LAMBDA        128            /* the LAMBDA column of the NFA array */
#define LAMBDA_STAR   129            /* the LAMBDA* column of the NFA array */
#define DELTA         128            /* the start state column of the DFA */

/*---------------------------------------------------  input/output constants */
#define COMMENT       "###"          /* for delimiting comments */
#define MACHINE       "!name:"       /* for delimiting string w/ name */
#define SIGMA         "!alphabet:"   /* for delimiting string w/ alphabet */
#define START         "!start:"      /* for delimiting string w/ start states */
#define ACCEPT        "!accept:"     /* for delimiting string w/ accept states*/

/*------------------ F U N C T I O N   P R O T O T Y P E S -------------------*/

   void bit_print ( unsigned int a);
/* displays bits for me */

   void nfa_init(unsigned int N[MAX_STATES + 1][SIGMA_STAR + 2]);
/* initializes the nfa */

   void dfa_init(unsigned int D[2^MAX_STATES + 1][SIGMA_STAR + 1]);
/* initializes the dfa */

   void variable_init(void);
/* initializes the numerous variables  for another NFA-DFA construction */

   void nfa_display(unsigned int N[MAX_STATES + 1][SIGMA_STAR + 2], unsigned int H);
/* displays the non empty states of the nfa */

   void dfa_display(unsigned int D[2^MAX_STATES + 1][SIGMA_STAR + 1], unsigned int H);
/* displays the non empty states of the dfa */

   unsigned int count_states(unsigned int cur_set);
/* takes an unsigned integer representing the set of states you want counted
   - returns the number of states in that set */

   void construct_dfa(unsigned int N[MAX_STATES + 1][SIGMA_STAR + 2], unsigned int start_set,
                     unsigned int accept_set,  unsigned int high_state);
/*  Takes a filled nfa array, an unsigned integer for its start states, its accept states,
    the highest state; creates a dfa using the global dfa variables since we
    will only have one dfa at a time.  Calls the dfa_init.  */

   void output_data(FILE **fp, unsigned int D[2^MAX_STATES + 1][SIGMA_STAR + 1],
                   unsigned int start_set, unsigned int accept_set, unsigned int high_state);
/*  Takes a filled dfa array, an unsigned integer for its start states, its accept states,
    the highest state and a pointer to a pointer to the file;
    outputs the dfa specs to the file. */

   void output_sets(FILE **fp, unsigned int d[MAX_STATES+ 1],
                   unsigned int high_state);
/*  Takes the array with the state to set mapping for the original
    DFA before reversal, and prints it to the output file. */

   unsigned int menu_options(void);
/*  Promts the user for the option of performing 2 Regular-preserving Operations
    (Compliment and Reverse) on the recently constructed dfa.  Returns the choice. */

   void reverse_transitions(unsigned int D[2^MAX_STATES + 1][SIGMA_STAR + 1],
                     unsigned int high_state);
/*  Takes a reduced DFA and reverses the transitions for each alphabet
    on each reachable state. Stores this information in a global reverse NFA array. */
/*--------------------------------------------------- my numerous variables!! */

   FILE  *infp,                             /* pointer to input file */
   *tempfp,                           /* pointer to my temp file */
   *outfp,                            /* pointer to output file */
   *newfp,
   *fopen();                          /* pointer to function */

   char  SIGMA_DELM[] = "!alphabet: \n",    /* delimiter for the alphabet */
   MACHINE_DELM[] = "!name: \n",      /* delimiter for the name */
   A_DELM[] = "\"",                   /* delimiter for the alphabet */
   S_DELM[] = " ,;\n",                /* delimiter for the start states */
   D_DELM[] = ":",                    /* delimiter for the declaration line*/
   AL_DELM[] = "\'",                  /* delimiter for the alphabet */
   T_DELM[] = "=>; ,\n",              /* delimiter for the transition */
   NEWFILE[] = "extra.dat",
   
   *name,          /* pointer to name - to be malloced later */
   *name_temp,     /* pointer to name - for strtok */
   
   *sigma,         /* pointer to alphabet - to be malloced later */
   *sigma_temp,    /* pointer to alphabet - for strtok */
   sigma_array[SIGMA_STAR], /* array for easy accessing of the alphabet */
   
   *state_temp,    /* pointer to state - for strtok */
   *start_temp,    /* pointer to start states- for strtok */
   *accept_temp,   /* pointer to accept states - for strtok */
   *trans_temp,    /* pointer to transition - for strtok */
   trans_decl[MAX_STATES][INP_LENGTH + 1],  /* for transition decarations */
   
   inp_line[INP_LENGTH + 1],    /* input line */
   temp_line[INP_LENGTH + 1],   /* for use with strtok */
   brute_line[INP_LENGTH + 1],  /* brute force way of saving inp_line */
   
   *part,                      /* pointer to part of string - for strtok */
   string_line[MAX_STATES*2],  /* string line for output */
   string_part[MAX_STATES*2];  /* array of string parts for output */



   unsigned int  i, j, k, x, y, z, l, max,          /* for incrementation and for loops */
   choice,
                              /* for menu selection */
   options,
   options_flag = FALSE,
   
   states[MAX_STATES + 1] =
   { 0x0,          /*the bit representation of all the states*/
     0x1,0x2,0x4,0x8,
     0x10,0x20,0x40,0x80,
     0x100,0x200,0x400,0x800,
     0x1000,0x2000,0x4000,0x8000,
     0x10000,0x20000,0x40000,0x80000,
     0x100000,0x200000,0x400000,0x800000,
     0x1000000,0x2000000,0x4000000,0x8000000,
     0x10000000,0x20000000,0x40000000,0x80000000},
   
   sigma_char,          /* symbol - as an integer for ascii array indexing */
   sigma_count,         /* for size of alphabet */
   start_count,         /* for number of start states */
   accept_count,        /* for number of accept states */
   machine_count,       /* the number of machines given in the input */
   totalcom_count[10],
   compliment_count[10],  /* array to determine if complimented an even # */
   reverse_count[10],     /* array to determine # of times reversed */
   comrev_count[10],     /* array to determine # of times compliment of reverse */
   revcom_count[10],     /* array to determine # of times reverse of compliment*/
   been_reversed,       /* boolean - true if the dfa has been reversed */
   
   set_count,           /* the number of states in a set */
   count,               /* arbitrary count for comparison of sets */
   nfa_start_count,   /* the number of start states in NFA */
   dfa_start_count,   /* the number of start states in DFA */
   dfa_accept_count,  /* the number of accept states in DFA */
   
   highest_nfa_state, /* maximum states achieved in NFA */
   highest_dfa_state, /* maximum states reachable in DFA */
   highest_rev_state, /* maximum states reachable in reverse DFA */
   
   is_dfa_accept,       /* boolean to flag an accept state in dfa*/
   new_state,           /* new state created in dfa construction */
   existing_state,      /* boolean to flag an existing state in dfa*/
   
   state,               /* single state */
   start_state,         /* single start state */
   accept_state,        /* single accept state */
   trans_state,         /* single transition state */
   
   dfa_start_set,   /* the bit representation of dfa start state set */
   nfa_start_set,   /* the bit representation of nfa start state set */
   rev_start_set,   /* the bit representation of reverse start state set */
   
   dfa_accept_set,  /* the bit representation of dfa accept state set */
   nfa_accept_set,  /* the bit representation of nfa accept state set */
   rev_accept_set,  /* the bit representation of reverse accept state set */
   com_accept_set,  /*the bit representation of compliment accept state set*/
   
   delta_array[MAX_STATES + 1],
   nfa[MAX_STATES + 1][SIGMA_STAR + 2],    /* the NFA 2D array */
   dfa[2^MAX_STATES +1][SIGMA_STAR + 1],   /* the DFA 2d array */
   reverse_nfa[MAX_STATES + 1][SIGMA_STAR + 2];/* the reverse 2D array */

/*--------------------------  B E G I N  M A I N  ----------------------------*/
   int main (int argc, char ** argv)
   {

   
      if(argc < 2)   /* opens input file */
      {
         fprintf(stderr, "\nYou must enter 2 command-line arguments following ");
         fprintf(stderr, "the executable:\n\n");
         fprintf(stderr,
                "C:\\> program_executable   your_input_file.dat   your_output_file.dat\n");
         exit(1);
      };
   
   /*-----------------------------------------------------------------open files */
      if((infp = fopen(argv[1], "r")) == NULL)   /* opens input file */
      {
         fprintf(stderr, "Error opening input file.");
         exit(1);
      };
   
   
      if((outfp = fopen(argv[2], "w")) == NULL)   /* opens output file */
      {
         fprintf(stderr, "Error opening output file.");
         exit(1);
      };
   
   /*------------------------------------------------------ initialize variables */
      variable_init();
   /*------------------------------------------------- fill nfa w/ data from file*/
      fgets(inp_line, INP_LENGTH, infp);
   
      while (feof(infp)== FALSE)
      {
         while((inp_line[0] == '#') || (inp_line[0] == '\n'))
            fgets(inp_line, INP_LENGTH, infp);
         while((inp_line[0] != '#') &&  (feof(infp)== FALSE))
         {
         
            if (inp_line[0] != '\n')
            {
               if ((inp_line[0] != '#') && (inp_line[0] != '!'))
               {                               /* if a transition declaration */
                  for (i = 0; i <  MAX_STATES; i++)
                  {                             /* initialize trans_decl holder */
                     trans_decl[i][0] = '\0';
                  }
               
                  state_temp = strtok(inp_line, D_DELM);    /* get state */
                  state = atoi(state_temp);
                  if (state > highest_nfa_state)
                     highest_nfa_state = state;
                  if (state != 0)                     /* to avoid strtok on an empty line */
                  {                                        /* get transition declarations */
                     strcpy(temp_line, inp_line + (strlen(state_temp) + 1));
                     i = 0;
                     trans_temp = strtok(temp_line, ";\n");
                     strcpy(trans_decl[i], trans_temp);
                     while((trans_temp = strtok(NULL, ";\n")) != NULL)
                     {
                        i++;
                        strcpy(trans_decl[i], trans_temp);
                     }
                  
                     for (j = 0; j < i+1; j++)        /* parse each transition declaration */
                     {
                        k = 0;
                        while (trans_decl[j][k] == ' ')        /* eliminate leading blanks */
                           k++;
                        strcpy(trans_decl[j], trans_decl[j] + k);
                     
                        if (trans_decl[j][0] != '=')             /* if an alphabet symbol */
                        {
                           sigma_temp = strtok(trans_decl[j], AL_DELM);
                           sigma_char = *sigma_temp;
                           while((state_temp = strtok(NULL, T_DELM)) != NULL)
                           {
                              trans_state = atoi(state_temp);
                              nfa[state][sigma_char] =
                                 nfa[state][sigma_char] | states[trans_state];
                           }
                        }/* end if */
                        else                                              /*  else LAMBDA */
                        {
                           state_temp = strtok(trans_decl[j], T_DELM);
                           trans_state = atoi(state_temp);
                           nfa[state][LAMBDA] = nfa[state][LAMBDA] | states[trans_state];
                           while((state_temp = strtok(NULL, T_DELM)) != NULL)
                           {
                              trans_state = atoi(state_temp);
                              nfa[state][LAMBDA] = nfa[state][LAMBDA] | states[trans_state];
                           }
                        }/* end else */
                     } /* end for */
                  }/* end if state not 0 */
               }
               else if (inp_line[0] == '!')                     /* if a specification */
               {
                  if (inp_line[1] == 'n')
                  {                                                     /* if the name */
                     name_temp = strtok(inp_line, MACHINE_DELM);
                     name = (char *)malloc(strlen(name_temp) + 1 * sizeof(char));
                     strcpy(name, name_temp);
                  }
                  else if  (inp_line[2] == 'l')
                  {                                                /* if the alphabet */
                     sigma_temp = strtok(inp_line, A_DELM);
                     sigma_temp = strtok(NULL, A_DELM);
                     sigma_count = strlen(sigma_temp);
                     sigma = (char *)malloc(sigma_count + 1 * sizeof(char));
                     strcpy(sigma, sigma_temp);
                     for (k = 0; k < sigma_count; k++)
                     {
                        sigma_array[sigma[k]] = '1';
                     }
                  }
                  else if  (inp_line[1] == 's')
                  {                                            /* if the start states */
                     start_temp = strtok(inp_line, S_DELM);
                     while((start_temp = strtok(NULL, S_DELM)) != NULL)
                     {
                        start_count++;
                        start_state = atoi(start_temp);
                        nfa_start_set = nfa_start_set | states[start_state];
                     }
                  #ifdef DEBUG
                  printf("the start states:\n");
                  bit_print(nfa_start_set);
                  #endif
                  }
                  else if  (inp_line[2] == 'c')
                  {                                         /*   if the accept states  */
                     accept_temp = strtok(inp_line, S_DELM);
                     while((accept_temp = strtok(NULL, S_DELM)) != NULL)
                     {
                        accept_count++;
                        accept_state = atoi(accept_temp);
                        nfa_accept_set = nfa_accept_set | states[accept_state];
                     }
                  #ifdef DEBUG
                  printf("the accept states:\n");
                  bit_print(nfa_accept_set);
                  #endif
                  }
               }
            }/* end if not a blank line '\n' */
            fgets(inp_line, INP_LENGTH, infp);
            strcpy(brute_line, inp_line);
         } /* end while not a comment */
      
         construct_dfa(nfa, nfa_start_set, nfa_accept_set, highest_nfa_state);
      
         machine_count++;/* completed a conversion */
      
      /*-------------------------------------------------  generate output for file */
      
         fprintf(outfp,"%s This is the output of the DFA constructed from the NFA, %s\n",
                COMMENT, name);
         output_data(&outfp, dfa, dfa_start_set, dfa_accept_set,
                    highest_dfa_state);
      
      
      /*----------------------------------------------------- free pointers & space */
         free(name);
         free(sigma);
      
         variable_init();
      
      }/* end while not end of file */
   /*--------------------------------------------------------------  close files */
      fclose(outfp);
   /*-------------------------------------------------  terminal output for user */
      printf("\n\n\tYour input file, \"%s\" has been successfully read. \n\n",
            argv[1]);
      if (machine_count > 1)
      {                       /* if more than one machine in input */
         printf("You submitted %d NFA machines in your file.\n", machine_count);
         printf("The equivalent DFAs, in their reduced form, have been constructed.\n\n");
      }
      else
      {
         printf("You submitted %d NFA machine in your file.\n", machine_count);
         printf("The equivalent DFA, in it\'s reduced form, has been constructed.\n\n");
      }
      printf("\tThe output file, \"%s\", has been generated.\n\n", argv[2]);
   
   /*------------------------------------------------------E X T R A--------------*/
   
      printf("\nYou now have the option to rewind the file and perform \"Regular-Preserving\"\n");
      printf("operations on the constructed DFA's.\n\n");
      printf("If you choose to continue, a new file - \"%s\" - will be generated\n", NEWFILE);
      printf("with this extra data in the same format as the assignment file.\n\n");
      printf("Please type the number corresponding to your choice:  \n\n");
      printf("\t(%s)  %s\n","1" , "Generate a new file with \"Regular-Preserving\" DFA's.\n\n");
   
      printf("\t(%s)  %s\n","2" , "End this Program");
      do {
         scanf("%d", &options);
         if((options < 1) || (options > 2))
            printf("You need to enter a 1 or a 2:");
      } while ((options < 1) || (options > 2));
   
      if(options == 1)          /* if you want to apply some regular-preserving operations */
      {
         options_flag = TRUE;
         rewind(infp);
         machine_count = 0;
         if((newfp = fopen(NEWFILE, "w")) == NULL)   /* opens output file */
         {
            fprintf(stderr, "Error opening new file.");
            exit(1);
         };
      /*------------------------------------------------------ initialize variables */
         variable_init();
      /*------------------------------------------------- fill nfa w/ data from file*/
         fgets(inp_line, INP_LENGTH, infp);
      
         while (feof(infp)== FALSE)
         {
            while((inp_line[0] == '#') || (inp_line[0] == '\n'))
               fgets(inp_line, INP_LENGTH, infp);
            while((inp_line[0] != '#') &&  (feof(infp)== FALSE))
            {
            
               if (inp_line[0] != '\n')
               {
                  if ((inp_line[0] != '#') && (inp_line[0] != '!'))
                  {                               /* if a transition declaration */
                     for (i = 0; i <  MAX_STATES; i++)
                     {                             /* initialize trans_decl holder */
                        trans_decl[i][0] = '\0';
                     }
                  
                     state_temp = strtok(inp_line, D_DELM);    /* get state */
                     state = atoi(state_temp);
                     if (state > highest_nfa_state)
                        highest_nfa_state = state;
                     if (state != 0)                     /* to avoid strtok on an empty line */
                     {                                        /* get transition declarations */
                        strcpy(temp_line, inp_line + (strlen(state_temp) + 1));
                        i = 0;
                        trans_temp = strtok(temp_line, ";\n");
                        strcpy(trans_decl[i], trans_temp);
                        while((trans_temp = strtok(NULL, ";\n")) != NULL)
                        {
                           i++;
                           strcpy(trans_decl[i], trans_temp);
                        }
                     
                        for (j = 0; j < i+1; j++)        /* parse each transition declaration */
                        {
                           k = 0;
                           while (trans_decl[j][k] == ' ')        /* eliminate leading blanks */
                              k++;
                           strcpy(trans_decl[j], trans_decl[j] + k);
                        
                           if (trans_decl[j][0] != '=')             /* if an alphabet symbol */
                           {
                              sigma_temp = strtok(trans_decl[j], AL_DELM);
                              sigma_char = *sigma_temp;
                              while((state_temp = strtok(NULL, T_DELM)) != NULL)
                              {
                                 trans_state = atoi(state_temp);
                                 nfa[state][sigma_char] =
                                    nfa[state][sigma_char] | states[trans_state];
                              }
                           }/* end if */
                           else                                              /*  else LAMBDA */
                           {
                              state_temp = strtok(trans_decl[j], T_DELM);
                              trans_state = atoi(state_temp);
                              nfa[state][LAMBDA] = nfa[state][LAMBDA] | states[trans_state];
                              while((state_temp = strtok(NULL, T_DELM)) != NULL)
                              {
                                 trans_state = atoi(state_temp);
                                 nfa[state][LAMBDA] = nfa[state][LAMBDA] | states[trans_state];
                              }
                           }/* end else */
                        } /* end for */
                     }/* end if state not 0 */
                  }
                  else if (inp_line[0] == '!')                     /* if a specification */
                  {
                     if (inp_line[1] == 'n')
                     {                                                     /* if the name */
                        name_temp = strtok(inp_line, MACHINE_DELM);
                        name = (char *)malloc(strlen(name_temp) + 1 * sizeof(char));
                        strcpy(name, name_temp);
                     }
                     else if  (inp_line[2] == 'l')
                     {                                                /* if the alphabet */
                        sigma_temp = strtok(inp_line, A_DELM);
                        sigma_temp = strtok(NULL, A_DELM);
                        sigma_count = strlen(sigma_temp);
                        sigma = (char *)malloc(sigma_count + 1 * sizeof(char));
                        strcpy(sigma, sigma_temp);
                        for (k = 0; k < sigma_count; k++)
                        {
                           sigma_array[sigma[k]] = '1';
                        }
                     }
                     else if  (inp_line[1] == 's')
                     {                                            /* if the start states */
                        start_temp = strtok(inp_line, S_DELM);
                        while((start_temp = strtok(NULL, S_DELM)) != NULL)
                        {
                           start_count++;
                           start_state = atoi(start_temp);
                           nfa_start_set = nfa_start_set | states[start_state];
                        }
                     #ifdef DEBUG
                     printf("the start states:\n");
                     bit_print(nfa_start_set);
                     #endif
                     }
                     else if  (inp_line[2] == 'c')
                     {                                         /*   if the accept states  */
                        accept_temp = strtok(inp_line, S_DELM);
                        while((accept_temp = strtok(NULL, S_DELM)) != NULL)
                        {
                           accept_count++;
                           accept_state = atoi(accept_temp);
                           nfa_accept_set = nfa_accept_set | states[accept_state];
                        }
                     #ifdef DEBUG
                     printf("the accept states:\n");
                     bit_print(nfa_accept_set);
                     #endif
                     }
                  }
               }/* end if not a blank line '\n' */
               fgets(inp_line, INP_LENGTH, infp);
               strcpy(brute_line, inp_line);
            } /* end while not a comment */
         
            construct_dfa(nfa, nfa_start_set, nfa_accept_set, highest_nfa_state);
         
            machine_count++;            /* completed a conversion */
         
         /*-------------------------------------------------  generate output for file */
         
            fprintf(outfp,"%s This is the output of the DFA constructed from the NFA, %s\n",
                   COMMENT, name);
            output_data(&newfp, dfa, dfa_start_set, dfa_accept_set,
                       highest_dfa_state);
         
            if ((machine_count == 1) || (machine_count == 21))
            {
               printf("\n\nThe %dst NFA machine has been read from your input file.\n",
                     machine_count);
               printf("The equivalent DFA, in it\'s reduced form, has been constructed.\n");
               printf("And the data has been generated to the new file, \"%s\".\n\n",
                     NEWFILE);
            }
            else if ((machine_count == 2) || (machine_count == 22))
            {
               printf("\n\nThe %dnd NFA machine has been read from your input file.\n",
                     machine_count);
               printf("The equivalent DFA, in it\'s reduced form, has been constructed.\n");
               printf("And the data has been generated to the new file, \"%s\".\n\n",
                     NEWFILE);
            }
            else if ((machine_count == 3) || (machine_count == 33))
            {
               printf("\n\nThe %drd NFA machine has been read from your input file.\n",
                     machine_count);
               printf("The equivalent DFA, in it\'s reduced form, has been constructed.\n");
               printf("And the data has been generated to the new file, \"%s\".\n\n",
                     NEWFILE);
            }
            else
            {
               printf("\n\nThe %dth NFA machine has been read from your input file.\n",
                     machine_count);
               printf("The equivalent DFA, in it\'s reduced form, has been constructed.\n");
               printf("And the data has been generated to the new file, \"%s\".\n\n",
                     NEWFILE);
            }
         
            choice = menu_options();
            been_reversed = FALSE;
         /*--------------------------------------------------  offer user menu options */
            while (choice < QUIT)
            {
            /*--------------------------------------------------  compliment option */
            
               if (choice == COMPLIMENT)
               {                     /* take the compliment of the accept states */
               
                  com_accept_set = CAP_PHI;
                  for (j = 1; j <= highest_dfa_state; j++)
                  {
                     if ((dfa_accept_set & states[j]) == states[j])
                     {
                        com_accept_set = ~states[j];
                        dfa_accept_set = dfa_accept_set & com_accept_set;
                     }
                     else
                     {
                        dfa_accept_set = dfa_accept_set | states[j];
                     
                     }
                  } /* end take the compliment  */
               
                  fprintf(outfp,"%s This is the above DFA - Complimented.\n", COMMENT);
                  if (been_reversed == FALSE)
                  {
                     compliment_count[machine_count]++;
                  }
                  if ((compliment_count[machine_count] > 1) && (been_reversed == FALSE))
                  {
                     fprintf(newfp,"%s This is the original DFA - Complimented %d time(s).\n",
                            COMMENT, compliment_count[machine_count]);
                     if ((compliment_count[machine_count] % 2) == 0)
                     {
                        printf("\nYou have Complimented the original machine %d time(s).\n",
                              compliment_count[machine_count]);
                        printf("It is now back to the original DFA machine generated ");
                        printf("from your NFA input.");
                     }
                  }
                  totalcom_count[machine_count]++;
               
                  output_data(&newfp, dfa, dfa_start_set, dfa_accept_set,
                             highest_dfa_state);
                  printf("\n\n\nThe Compliment DFA has been constructed.\n");
                  printf("And the data has been generated to the new file, \"%s\".\n\n",
                        NEWFILE);
                  choice = menu_options();
               }
               /*--------------------------------------------------  reverse option */
               else if (choice == REVERSE)
               {
               
                  been_reversed = TRUE;
                  reverse_count[machine_count]++;
               
                  rev_start_set = dfa_accept_set;
                  rev_accept_set = dfa_start_set;
               
                  for (j = 1; j <= (MAX_STATES); j++)
                  {                             /*initialize set states of delta */
                     delta_array[j] =  0;
                  }
               
                  highest_rev_state = highest_dfa_state;
               
                  for (j = 1; j <= highest_rev_state; j++)
                  {                             /* copy the set states to delta */
                     delta_array[j] =  dfa[j][DELTA];
                  }
               
                  reverse_transitions(dfa, highest_rev_state);
               
                  printf("%d", highest_rev_state);
               
                  construct_dfa(reverse_nfa, rev_start_set, rev_accept_set,
                               highest_rev_state);
               
                  fprintf(newfp,"%s This is the above DFA - Reversed.\n", COMMENT);
               
                  output_data(&newfp, dfa, dfa_start_set, dfa_accept_set,
                             highest_dfa_state);
                  output_sets(&newfp, delta_array, highest_rev_state);
               
                  printf("\n\n\nThe Reverse DFA has been constructed.\n");
                  printf("And the data has been generated to the new file, \"%s\".\n\n",
                        NEWFILE);
                  choice = menu_options();
               }
               else if (choice == COMPLIMENT_OF_REVERSE)
               {
                  been_reversed = TRUE;
                  comrev_count[machine_count]++;
               
                  rev_start_set = dfa_accept_set;
                  rev_accept_set = dfa_start_set;
               
                  for (j = 1; j <= (MAX_STATES); j++)
                  {                             /* copy the set states to delta */
                     delta_array[j] =  0;
                  }
               
                  highest_rev_state = highest_dfa_state;
                  for (j = 1; j <= highest_rev_state; j++)
                  {                             /* copy the set states to delta */
                     delta_array[j] =  dfa[j][DELTA];
                  }
               
                  reverse_transitions(dfa, highest_rev_state);
               
                  construct_dfa(reverse_nfa, rev_start_set, rev_accept_set,
                               highest_rev_state);
               
                   /* take the compliment of the accept states */
                  com_accept_set = CAP_PHI;
                  for (j = 1; j <= highest_dfa_state; j++)
                  {
                     if ((dfa_accept_set & states[j]) == states[j])
                     {
                        com_accept_set = ~states[j];
                        dfa_accept_set = dfa_accept_set & com_accept_set;
                     }
                     else
                     {
                        dfa_accept_set = dfa_accept_set | states[j];
                     
                     }
                  } /* end take the compliment  */
                  fprintf(outfp,"%s This is the Compliment of the Reverse of the above DFA.\n", COMMENT);
               
                  output_data(&outfp, dfa, dfa_start_set, dfa_accept_set,
                             highest_dfa_state);
               
                  printf("\n\n\nThe Compliment of the Reverse DFA has been constructed.\n");
                  printf("And the data has been generated to the new file, \"%s\".\n\n",
                        NEWFILE);
                  choice = menu_options();
               }
               else if (choice == REVERSE_OF_COMPLIMENT)
               {
                   /* take the compliment of the accept states */
                  revcom_count[machine_count]++;    
                  com_accept_set = CAP_PHI;
                  for (j = 1; j <= highest_dfa_state; j++)
                  {
                     if ((dfa_accept_set & states[j]) == states[j])
                     {
                        com_accept_set = ~states[j];
                        dfa_accept_set = dfa_accept_set & com_accept_set;
                     }
                     else
                     {
                        dfa_accept_set = dfa_accept_set | states[j];
                     
                     }
                  } /* end take the compliment  */
                  been_reversed = TRUE;
                  rev_start_set = dfa_accept_set;
                  rev_accept_set = dfa_start_set;
               
                  for (j = 1; j <= (MAX_STATES); j++)
                  {                             /* copy the set states to delta */
                     delta_array[j] =  0;
                  }
               
                  highest_rev_state = highest_dfa_state;
                  for (j = 1; j <= highest_rev_state; j++)
                  {                             /* copy the set states to delta */
                     delta_array[j] =  dfa[j][DELTA];
                  }
               
                  reverse_transitions(dfa, highest_rev_state);
               
                  construct_dfa(reverse_nfa, rev_start_set, rev_accept_set,
                               highest_rev_state);
               
                  fprintf(outfp,"%s This is the Reverse of the Compliment of the above DFA.\n", COMMENT);
               
                  output_data(&outfp, dfa, dfa_start_set, dfa_accept_set,
                             highest_dfa_state);
                  printf("\n\n\nThe Reverse of the Compliment DFA has been constructed.\n");
                  printf("And the data has been generated to the new file, \"%s\".\n\n",
                        NEWFILE);
                  choice = menu_options();
               }
            } /* end while not QUIT */
         
         
         
         /*----------------------------------------------------- free pointers & space */
            free(name);
            free(sigma);
         
            variable_init();
         
         }/* end while not end of file */
      } /* end if you want to apply some regular-preserving operations */
   /*--------------------------------------------------------------  close files */
      fclose(infp);
   /*-------------------------------------------------  terminal output for user */
   
      printf("\n\n\tThank you for your time.\n\n");
      if (options_flag)
      {
         if (machine_count > 1)
         {                       /* if more than one machine in input */
            printf("You submitted %d NFA machines in your file.\n", machine_count);
            printf("The equivalent DFAs, in their reduced form, have been constructed.\n");
            printf("And the following operations have also been constructed.\n\n");
            printf("%s    %s    %s    %s    %s\n\n", "Machine", "Compliments", "Reverses",
                  "Comp. of Reverse", "Reverse of the Comp.");
         
            for (j = 1; j <= machine_count; j++)
            {
               printf("%5d%13d%13d%15d%18d\n\n", j, totalcom_count[j], reverse_count[j],
                     comrev_count[j], revcom_count[j]);
            
            }
         }
         else
         {
            printf("You submitted %d NFA machine in your file.\n", machine_count);
            printf("The equivalent DFA, in it\'s reduced form, has been constructed.\n");
            printf("And the following operations have also been constructed.\n\n");
            printf("%s    %s    %s    %s    %s\n\n", "Machine", "Compliments", "Reverses",
                  "Comp. of Reverse", "Reverse of the Comp.");
            for (j = 1; j <= machine_count; j++)
            {
               printf("%5d%13d%13d%15d%18d\n\n", j, totalcom_count[j], reverse_count[j],
                     comrev_count[j], revcom_count[j]);
            
            }
         
         }
         printf("The output file, \"%s\", has been generated.\n\n", NEWFILE);
      
      
      }
   
      fclose(newfp);
      return EXIT_SUCCESS;
   }


/*----------------------------  F U N C T I O N S  ---------------------------*/


/*---------------------------  N F A _ I N I T   ------------------------------\
    initializes the nfa
\-----------------------------------------------------------------------------*/
   void nfa_init(unsigned int N[MAX_STATES + 1][SIGMA_STAR + 2])
   {
      unsigned int i, j;
      for(i = 0; i < MAX_STATES + 1; i++)
      {
         for (j = 0; j < SIGMA_STAR + 2; j++)
         {
            N[i][j] = CAP_PHI;
         }
      }
   } /* end nfa_init */

/*---------------------------  D F A _ I N I T   ------------------------------\
    initializes the dfa
\-----------------------------------------------------------------------------*/
   void dfa_init(unsigned int D[2^MAX_STATES + 1][SIGMA_STAR + 1])
   {
      unsigned int i, j;
      for(i = 0; i <= (2^MAX_STATES); i++)
      {
         for (j = 0; j < SIGMA_STAR + 1; j++)
         {
            D[i][j] = CAP_PHI;
         }
      }
   } /* end dfa_init */

/*------------------------  V A R I A B L E _ I N I T  ------------------------\
    initializes the numerous variables  for another NFA-DFA construction
\-----------------------------------------------------------------------------*/
   void variable_init(void)
   {
      unsigned int k = 0;
   
      trans_decl[MAX_STATES][INP_LENGTH + 1] = '\0';  /* for transitions */
      temp_line[INP_LENGTH + 1] = '\0';   /* for use with strtok */
   
      sigma_char = 0;         /* symbol - as an integer for ascii array indexing */
      sigma_count = 0;          /* for size of alphabet */
      start_count = 0;          /* for number of start states */
      accept_count = 0;        /* for number of accept states */
   
      been_reversed = FALSE;       /* boolean - true if the dfa has been reversed */
   
      set_count = CAP_PHI;           /* the number of states in a set */
      count = CAP_PHI;               /* arbitrary count for comparison of sets */
      nfa_start_count = CAP_PHI;   /* the number of start states in NFA */
      dfa_start_count = CAP_PHI;   /* the number of start states in DFA */
      dfa_accept_count = CAP_PHI;  /* the number of accept states in DFA */
   
      highest_nfa_state = CAP_PHI; /* maximum states achieved in NFA*/
      highest_dfa_state = CAP_PHI; /* maximum states reachable in DFA */
   
      is_dfa_accept = FALSE;       /* boolean to flag an accept state in dfa*/
      new_state = CAP_PHI;           /* new state created in dfa construction */
      existing_state = FALSE;      /* boolean to flag an existing state in dfa*/
   
      state = CAP_PHI;               /* single state */
      start_state = CAP_PHI;         /* single start state */
      accept_state = CAP_PHI;        /* single accept state */
      trans_state = CAP_PHI;         /* single transition state */
   
      dfa_start_set = CAP_PHI;   /* the bit representation of dfa start state set */
      nfa_start_set = CAP_PHI;   /* the bit representation of nfa start state set */
      rev_start_set = CAP_PHI;   /* the bit representation of reverse start state set */
   
      dfa_accept_set = CAP_PHI;  /* dfa accept state set */
      nfa_accept_set = CAP_PHI;  /* nfa accept state set */
      rev_accept_set = CAP_PHI;  /* reverse accept state set */
      com_accept_set = CAP_PHI;/*compliment accept state set*/
   
   
      trans_decl[MAX_STATES][INP_LENGTH + 1] = '\0';  /* for transitions */
      temp_line[INP_LENGTH + 1] = '\0';   /* for use with strtok */
   
   
      sigma_count  = 0;            /* for size of alphabet */
      start_count  = 0;            /* for number of start states */
      accept_count = 0;            /* for number of accept states */
      set_count    = 0;            /* the number of states in a set */
      count        = 0;            /* arbitrary count for comparison of sets */
      nfa_start_count = CAP_PHI;   /* the number of start states in NFA */
      dfa_start_count = CAP_PHI;   /* the number of start states in DFA */
      dfa_accept_count = CAP_PHI;  /* the number of accept states in DFA */
      highest_nfa_state = CAP_PHI; /* maximum states achieved in NFA*/
      highest_dfa_state = CAP_PHI; /* maximum states reachable in DFA */
      dfa_start_set = CAP_PHI;     /* start state set for dfa */
      new_state     = CAP_PHI;     /* new state created in dfa construction */
      existing_state = FALSE;      /* boolean to flag an existing state in dfa*/
      is_dfa_accept  = FALSE;      /* boolean to flag an accept state in dfa*/
   
      nfa_start_set = CAP_PHI;  /* the bit representation of the start states */
      nfa_accept_set = CAP_PHI;/* the bit representation of nfa accept states */
      dfa_accept_set = CAP_PHI;/* the bit representation of dfa accept states */
   
      for (k = 0; k < SIGMA_STAR; k++)
      {
         sigma_array[k] = '0';
      }
      for (k = 0; k < 10; k++)
      {
         compliment_count[k] = FALSE;
      }
      nfa_init(nfa);
      nfa_init(reverse_nfa);
      dfa_init(dfa);
   
   } /* end variable_init */

/*---------------------------  N F A _ D I S P L A Y  -------------------------\
    displays the non empty states of the nfa
\-----------------------------------------------------------------------------*/
   void nfa_display(unsigned int N[MAX_STATES + 1][SIGMA_STAR + 2], unsigned int H)
   {
      unsigned int i, j;
      for(i = 0; i <= H; i++)
      {
         for (j = 0; j <= SIGMA_STAR + 2; j++)
         {
            if (N[i][j] != CAP_PHI)
            {
               if (j > 99)
               {
                  printf("nfa[%d][%d]:", i, j);
                  bit_print(N[i][j]);
               }
               else
               {
                  printf("nfa[%d][%d]:%s", i, j, BLANK);
                  bit_print(N[i][j]);
               }
            }
         }
      }
   } /* end nfa_display */

/*---------------------------  D F A _ D I S P L A Y  -------------------------\
    displays the non empty states of the dfa
\-----------------------------------------------------------------------------*/
   void dfa_display(unsigned int D[2^MAX_STATES + 1][SIGMA_STAR + 1], unsigned int H)
   {
      unsigned int i, j;
      for(i = 0; i <= H; i++)
      {
         for (j = 0; j <= SIGMA_STAR + 1; j++)
         {
            if (D[i][j] != CAP_PHI)
            {
               if (j > 99)
               {
                  printf("dfa[%d][%d]:", i, j);
                  bit_print(D[i][j]);
               }
               else
               {
                  printf("dfa[%d][%d]:%s", i, j, BLANK);
                  bit_print(D[i][j]);
               }
            }
         }
      }
   } /* end dfa_display */

/*--------------------------C O U N T _ S T A T E S  --------------------------\
    takes an integer representing the set of states you want counted
    - returns the number of states in that set
\-----------------------------------------------------------------------------*/
   unsigned int count_states(unsigned int cur_set)
   /*  */
   {
      unsigned int j, cnt = 0;
   
      for (j = 1; j <= MAX_STATES; j++)
      {
         if ((cur_set & states[j]) == states[j])
         {
            cnt++;
         }
      }
      return cnt;
   
   } /* end count_states */

/*----------------------- C O N S T R U C T _ D F A  --------------------------\
  Takes a filled nfa array, an integer for its start states, its accept states,
  the highest state; creates a dfa using the global dfa variables since we
  will only have one dfa at a time.  Calls the dfa_init.
\-----------------------------------------------------------------------------*/
   void construct_dfa(unsigned int N[MAX_STATES + 1][SIGMA_STAR + 2], unsigned int start_set,
                     unsigned int  accept_set,  unsigned int high_state)
   {
      dfa_init(dfa);
      dfa_start_set = CAP_PHI;
      dfa_accept_set = CAP_PHI;
      highest_dfa_state = CAP_PHI;
      new_state     = CAP_PHI;
      existing_state = FALSE;
      is_dfa_accept  = FALSE;
      dfa_accept_count = 0;
      set_count = 0;
      count = 0;     
   
      /*-------------------------------------------------------------  fill LAMBDA  */
   
      for (j = 1; j <= high_state; j++)
      {
         N[j][LAMBDA] = N[j][LAMBDA] | states[j];
      }
   
      /*----------------------------------------------------- calculate LAMBDA_STAR */
      for (i = 1; i <= high_state; i++)
      {
         N[i][LAMBDA_STAR] = N[i][LAMBDA_STAR] | N[i][LAMBDA];
         do {
            set_count = count_states(N[i][LAMBDA_STAR]);
            for (j = 1; j <= MAX_STATES; j++)
            {
               if ((N[i][LAMBDA_STAR] & states[j]) == states[j])
               {
                  N[i][LAMBDA_STAR] = N[i][LAMBDA_STAR] | N[j][LAMBDA];
               }
            } /* end for the states */
            count = count_states(N[i][LAMBDA_STAR]);
         } while (count > set_count);
      } /* end for each N state */
   
      /*---------------------------------------------- make start state for the dfa */
      dfa_start_set = dfa_start_set | start_set;
      for (j = 1; j <= MAX_STATES; j++)
      {
         if ((dfa_start_set & states[j]) == states[j])
         {
            dfa_start_set = dfa_start_set | N[j][LAMBDA_STAR];
         } /* end if state is a start state */
      }
      dfa_start_count = count_states(dfa_start_set);
      /*--------------------------------------------------------- calculate the dfa */
      new_state++;
      highest_dfa_state++;       /* set start state in the dfa */
      dfa[new_state][DELTA] = dfa[new_state][DELTA] | dfa_start_set;
      do{                        /* get the transitions for each alphabet */
         for (i = 0; i < SIGMA_STAR; i++)    /* for each alphabet */
         {
            if (sigma_array[i] == '1')  /* if i is in the alphabet */
            {      /* for each state in the dfa state set */
               for (j = 1; j <= MAX_STATES; j++)
               {                                  /* if state is in the set */
                  if((dfa[new_state][DELTA] & states[j]) == states[j])
                  {              /* union dfa nd nfa transditions for that alphabet */
                     dfa[new_state][i] = dfa[new_state][i] | N[j][i];
                  } /* end if state is in the set */
               }/* end for each state in the dfa state set */
               for (j = 1; j <= MAX_STATES; j++)
               {        /* now get LAMBDA_STAR for each state */
                  if((dfa[new_state][i] & states[j]) == states[j])
                  {              /* union dfa nd nfa transditions for that alphabet */
                     dfa[new_state][i] = dfa[new_state][i] | N[j][LAMBDA_STAR];
                  } /* end if state is in the set */
               } /* end get LAMBDA_STAR for each state */
               existing_state = FALSE;
            
               for (k = 1; k <= highest_dfa_state; k++)
               { /* check if new aplha set is equal to any state created in dfa */
                  if (dfa[k][DELTA] == dfa[new_state][i])
                  {
                     existing_state = TRUE;
                  }
               } /* end for each state created so far in dfa */
               if (existing_state == FALSE)
               {
                  highest_dfa_state++;
                  dfa[highest_dfa_state][DELTA] = dfa[new_state][i];
               }
            }/* end if */
         }/* end for each alphabet */
         new_state++;
      } while(new_state <= highest_dfa_state);
      /*-----------------------------------------------  find the dfa accept states */
      for (i = 1; i <= highest_dfa_state; i++)
      {
         is_dfa_accept = FALSE;
         for (j = 1; j <= MAX_STATES; j++)
         {
            if((accept_set & states[j])== states[j])
            {               /* if state is an accept state of the nfa */
               if ((dfa[i][DELTA] & states[j]) == states[j])
               {
                  is_dfa_accept = TRUE;
               }
            }/* end if */
         } /* end for j */
         if (is_dfa_accept)
         {
            dfa_accept_set = dfa_accept_set | states[i];
         } /* end if is_dfa_accept */
      } /* end for the dfa states */
      dfa_accept_count = count_states(dfa_accept_set);
   } /* end construct_dfa */

/*-------------------------- O U T P U T _ D A T A  ---------------------------\
  Takes a filled dfa array, an unsigned integer for its start states, its accept states,
  the highest state and a pointer to a pointer to the file;
  outputs the dfa specs to the file.
\-----------------------------------------------------------------------------*/
   void output_data(FILE **fp, unsigned int D[2^MAX_STATES + 1][SIGMA_STAR + 1],
                   unsigned int start_set, unsigned int accept_set, unsigned int high_state)
   {
      dfa_start_count = count_states(start_set);
      fprintf(*fp,"%s %s\n", MACHINE, name);
      fprintf(*fp,"%s \"%s\"\n", SIGMA, sigma);
      fprintf(*fp,"%s [", START);
      k = 0;
   
      /*-------------------------------------------------------  print start states */
      for (j = 1; j <= MAX_STATES; j++)
      {
         if ((start_set & states[j]) == states[j])
         {
            k++;
            if (k < dfa_start_count)
            {
               fprintf(*fp,"%d,", j);
            }
            else
            {
               fprintf(*fp,"%d];\n", j);
            }
         } /*end if */
      }/* end for start states */
      dfa_accept_count = count_states(dfa_accept_set);
      fprintf(*fp,"%s ", ACCEPT);
      k = 0;
      set_count = 0;
      /*------------------------------------------------------  print accept states */
      for (j = 1; j <= MAX_STATES; j++)
      {
         if ((accept_set & states[j]) == states[j])
         {
            set_count = count_states(D[j][DELTA]);
            k++;
            x = 0;
            if (k < dfa_accept_count)
            {
               fprintf(*fp,"[");
               for (i = 1; i <= MAX_STATES; i++)
               {
                  if ((D[j][DELTA] & states[i]) == states[i])
                  {
                     x++;
                     if (x < set_count)
                     {
                        fprintf(*fp,"%d,", i);
                     }
                     else
                     {
                        fprintf(*fp,"%d]", i);
                     }
                  }
               }
               if (set_count == 0)
               {
                  fprintf(*fp,"]");
               }
               fprintf(*fp,",");
            }
            else
            {
               fprintf(*fp,"[");
               for (i = 1; i <= MAX_STATES; i++)
               {
                  if ((D[j][DELTA] & states[i]) == states[i])
                  {
                     x++;
                     if (x < set_count)
                     {
                        fprintf(*fp,"%d,", i);
                     }
                     else
                     {
                        fprintf(*fp,"%d]", i);
                     }
                  }
               }
               if (set_count == 0)
               {
                  fprintf(*fp,"]");
               }
               fprintf(*fp,";\n");
            }
         } /*end if */
      }/* end for accept states */
   
      /*----------------------------------------------------- begin using temp file */
      if((tempfp=tmpfile()) == NULL)   /* opens my temp file */
      {
         fprintf(stderr, "Error opening output file.");
         exit(1);
      };
      /*-------------------------------------------- print transitions to temp file */
      for (j = 1; j <= high_state; j++)
      {
         x = 0;
         k = 0;
         set_count = count_states(D[j][DELTA]);
         fprintf(tempfp,"[");
         for (i = 1; i <= MAX_STATES; i++)
         {
            if ((D[j][DELTA] & states[i]) == states[i])
            {
               x++;
               if (x < set_count)
               {
                  fprintf(tempfp,"%d,", i);
               }
               else
               {
                  fprintf(tempfp,"%d] ", i);
               }
            }
         }
         if (x == 0)
         {
            fprintf(tempfp,"] ");
         }
         fprintf(tempfp,":");
         for (i = 0; i <= SIGMA_STAR; i++)
         {
            x =0;
            if (sigma_array[i] == '1')
            {
               fprintf(tempfp,"\'%c\'=>",i);
               k++;
               set_count = count_states(D[j][i]);
               fprintf(tempfp,"[");
               for (y = 1; y <= MAX_STATES; y++)
               {
                  if ((D[j][i] & states[y]) == states[y])
                  {
                     x++;
                     if (x < set_count)
                     {
                        fprintf(tempfp,"%d,", y);
                     }
                     else
                     {
                        fprintf(tempfp,"%d]; ", y);
                     }
                  }
               }
               if (x == 0)
               {
                  fprintf(tempfp,"]; ");
               }
            } /* end if */
         } /* end for */
         fprintf(tempfp,"\n");
      } /* end for transitions */
      /*---------------------------------------------- rewind temp file for reading */
      rewind(tempfp);
      l = 0; max = 0;
      /*----------------------------------  get the length of the longest DFA state */
      while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
      {
         strcpy(temp_line, string_line);
         part = strtok(temp_line, BLANK);
         l = strlen(part);
         if (l > max)
            max = l;
      }
      rewind(tempfp);
      /*---- thank goodness I'm coding this in the age of plentiful memory!!  -----*/
      /*---- these many if statements are the closest way I can figure to get
             them thar purty output that is required  -----*/
   
      if (max <= 5)
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-8s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-16s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      if ((max > 5) && (max <= 10))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-13s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-21s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      else if ((max > 10)&& (max <= 15))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-18s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-26s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      
      }
      else if ((max > 15) && (max <= 22))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-25s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-33s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      else if ((max > 22) && (max <= 30))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-33s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-41s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      else if ((max > 30) && (max <= 40))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-43s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-51s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      else if ((max > 40) && (max <= 50))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-53s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-61s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      else if ((max > 50) && (max <= 60))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-63s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-71s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      else if ((max > 60) && (max <= 70))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-73s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-81s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      else if ((max > 70) && (max <= 80))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-83s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-91s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      else if ((max > 80) && (max <= 90))
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-93s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-101s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
      else
      {
         while (fgets(string_line, INP_LENGTH, tempfp) != NULL)
         {
            strcpy(temp_line, string_line);
            part = strtok(temp_line, BLANK);
            strcpy(string_part, part);
            fprintf(*fp,"%-113s",string_part);
            while((part = strtok(NULL, " \n")) != NULL)
            {
               strcpy(string_part, part);
               fprintf(*fp,"%-111s", string_part);
            }
            fprintf(*fp,"\n");
         } /* end while get string */
      }
   
      fclose(tempfp);
      fprintf(*fp, "\n\n");
   } /* end output_data */


/*-------------------------- O U T P U T _ S E T S  ---------------------------\
  Takes the array with the state to set mapping for the original
  DFA before reversal, and prints it to the output file.
\-----------------------------------------------------------------------------*/
   void output_sets(FILE **fp, unsigned int d[MAX_STATES+ 1],
                   unsigned int high_state)
   {
      int i, k, x = 0;
      set_count = 0;
   
      fprintf(*fp,"%s This is the table relating the states of the Reversed DFA \n",
             COMMENT);
      fprintf(*fp,"%s to the Sets in the original DFA - before reversal.\n", COMMENT);
   
      for (k = 1; k <= high_state; k++)
      {
         x = 0;
         set_count = count_states(d[k]);
         fprintf(*fp,"%-4d", k);
         fprintf(*fp,":[");
         for (i = 1; i <= MAX_STATES; i++)
         {
            if ((d[k] & states[i]) == states[i])
            {
               x++;
               if (x < set_count)
               {
                  fprintf(*fp,"%d,", i);
               }
               else
               {
                  fprintf(*fp,"%d]", i);
               }
            }
         }
         if (set_count == 0)
         {
            fprintf(*fp,"]");
         }
         fprintf(*fp,"\n");
      }
      fprintf(*fp,"\n\n");
   } /* end output_sets */



/*-------------------------- M E N U _ O P T I O N S  -------------------------\
  Promts the user for the option of performing 2 Regular-preserving Operations
  (Compliment and Reverse) on the recently constructed dfa.  Returns the choice.
\-----------------------------------------------------------------------------*/
   unsigned int menu_options(void)
   {
      unsigned int C = 0;
   
      printf("\nYou have the option for the following \"Regular-Preserving\" operations:\n");
      printf("  (This menu applys ONLY to the most recently constructed DFA.)\n\n");
      printf("\t%-3d%s\n", COMPLIMENT, "Construct the Compliment DFA:");
      printf("\t%-3d%s\n", REVERSE, "Construct the Reverse DFA:");
      printf("\t%-3d%s\n", COMPLIMENT_OF_REVERSE,
            "Construct the Compliment of the Reverse DFA:"
            );
      printf("\t%-3d%s\n", REVERSE_OF_COMPLIMENT,
            "Construct the Reverse of the Compliment DFA:"
            );
      printf("\t%-3d%s\n\n", QUIT, "Quit this Menu:");
      printf("  (Choosing options 1 - 4 will generate the resulting DFA to the new file.)\n\n");
      printf("Please choose an option:  ");
      do {
         scanf("%d", &C);
         if((C > 5) || (C < 1) )
            printf("You need to enter a number between 1 and 5:");
      } while ((C > 5) || (C < 1));
      return C;
   } /* end menu_options */

/*-----------------  R E V E R S E _ T R A N S I T I O N S  -------------------\
  Takes a reduced DFA and reverses the transitions for each alphabet
  on each reachable state. Stores this information in a global reverse NFA array.
\-----------------------------------------------------------------------------*/
   void reverse_transitions(unsigned int D[2^MAX_STATES + 1][SIGMA_STAR + 1],
                     unsigned int h)
   {
      unsigned int i , j, k = 0;
   
      nfa_init(reverse_nfa);
   
      for (i = 0; i <= SIGMA_STAR; i++)
      {                      /* for each alphabet */
         if (sigma_array[i] == '1')
         {
            for (j = 1; j <= h; j++)
            {                      /* for each state in the dfa*/
               for (k = 1; k <= h; k++)
               {                        /*for each transition */
                  if (D[j][i]  == D[k][DELTA])
                  {
                     reverse_nfa[k][i] = reverse_nfa[k][i] | states[j];
                  }
               } /* end for each transition */
            } /* end for each state in the dfa*/
         } /* end if */
      } /* end for each alphabet */
   } /* end reverse_transitions */


/*---------------------------  B I T _ P R I N T ------------------------------\
    displays bits for me
\-----------------------------------------------------------------------------*/
   void bit_print ( unsigned int a)
   {
      unsigned int i;
      unsigned int n = sizeof( unsigned int) * CHAR_BIT;
      unsigned int mask = 1 << (n - 1);
      for (i = 1; i <= n; ++i)
      {
         putchar(((a & mask) == 0) ? '0' : '1');
         a <<=1;
         if (i % CHAR_BIT == 0 && i < n)
            putchar(' ');
      }
      printf("\n");
   } /* end bit_print */
