/****************************************************************************\
      Bettina "Goddess/Wonder Woman" McCormick 
  		March 31, 1998, COP 3402 - Systems, Assignment #2
  		Pass 1  - Due Date: March 31, 1998
 Version 3.141592653589793238462643383279502884- last updated  March 21, 1998
\****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/***********************    # D E F I N E S    ******************************/

/*-----------------------    OPERATION TABLE    ----------------------------*/

#define OP_LENGTH     6    	/* maximum number of chars in operation */
#define OT_SIZE     100    	/* max number of entries in Operation Table */

/*------------------------     SYMBOL TABLE    -----------------------------*/

#define HT_SIZE      59    	/* # of entries in the symtab_hash table */
#define STEP_SIZE    29    	/* value to be added to the symtab_hash value 
			           for resolving symtab_hash collisions */

/*-------------------------     ERROR QUEUE    -----------------------------*/

#define  INV_OP       0    	/* index to ERR_MSGS for invalid operation */
#define  DUP_LBL      1    	/* index to ERR_MSGS for duplicate label   */

/*------------------------     OPERAND QUEUE    ----------------------------*/

#define BLANK       ' '   	/* Defines a blank space (duh!) */
#define PLUS        '+'    	/* Defines a plus symbol (duh!) */
#define MINUS       '-'    	/* Defines a minus symbol (duh!) */
#define DELIM       " \t\n\0" /* Defines the different symbols which delimit 
				         strings */

/*---------------------------     PASS1    ---------------------------------*/

#define NAME_LENGTH  10    	/* maximum number of characters in name */
#define OPND_LENGTH  50    	/* maximum number of chars in operand */
#define INP_LENGTH   70    	/* max chars in an input assembly line */

/*-----------------------------   MAIN   -----------------------------------*/

/* #define DEBUG */         	/* Used for error checking */
#define TRUE         1     	/* Boolean value which indicates true */
#define FALSE        0     	/* Boolean value which indicates false */
#define NADA        -1    	/* Value used for empty SYMTAB slots */
#define RELOCATABLE 'R'    	/* Indicates a Relocatable address */
#define ABSOLUTE    'A'    	/* Indicates an Absolute address */

/***********************    S T R U C T U R E S    **************************/

/*-----------------------    OPERATION TABLE    ----------------------------*/
   struct ot_entry {       	/* an entry in Operation Table */
      char ot_mnemonic[OP_LENGTH + 1];
      int  ot_format;  /* 0 (assembler directive), 1,2,3 (formats 1,2,3/4) */
   };

   struct ot_entry optbl[OT_SIZE];  /* Operation Table; Operation Table is 
				               sorted with a binary search*/

   int 	  ot_count; 		/*number of entries in Operation Table */

/*------------------------     SYMBOL TABLE    -----------------------------*/

   struct st_entry {       	/* an entry in the Symbol Table */
      char st_name[NAME_LENGTH + 1];
      char st_type;        	/* R(elocatable) or A(bsolute) */
      int  st_value; 		/* value of the label */
   };

   struct st_entry symtbl[HT_SIZE]; /* Symbol Table; Symbol Table is maintained 
				       using hashing */

/*-------------------------     ERROR QUEUE    -----------------------------*/

   char  ERR_MSGS[2][25] = { "Invalid Operation",
                     "Duplicate Label" };  /* error messages */

   struct eq_entry {      	/* an entry in Error Queue */
      int     eq_lineno;  	/* line number */
      int     eq_msgind;  	/* index into ERR_MSGS(error message index) */
      struct  eq_entry  *eq_next; /* pointer to the next entry */
   };

   typedef struct eq_entry errnode;

   struct errtbl {          	 /* Error Queue; all the errors for the input 
   		      	    are maintained in a queue */  
      struct  eq_entry  *front; /* points to the front of the queue */
      struct  eq_entry  *rear;  /* points to the rear of the queue */
   };

/*------------------------     OPERAND QUEUE    ----------------------------*/

   struct opq_entry {   	/* an entry in Error Queue */
      char    opq_name[NAME_LENGTH + 1];
      int     opq_value;
      char    opq_type; 	/* R(elocatable) or A(bsolute) */
      char    opq_sign;  	/* (+)PLUS or (-)MINUS*/
      struct  opq_entry  *opq_next; /* pointer to the next entry */
   };

   typedef struct opq_entry opqnode;

   struct opndque {    		/* The Operand Queue; all the labels in the 
   			   expression are maintained in a queue */
      int qlength;              /* Holds size of the operand queue */
      struct  opq_entry  *front;/* points to the front of the queue */
      struct  opq_entry  *rear; /* points to the rear of the queue */ 
   };


/********************      D E C L A R A T I O N S    ***********************/

/*-----------------------------   MAIN   -----------------------------------*/

   FILE *ot_fptr,     /* File pointer used to open the operation table input*/
   *in_fptr,     /* File pointer used to open the assembly program */	
   *out_fptr,    /* File pointer used to open the output file */
   *fopen();     /* function for opening files */


   int  locctr;       /* location counter; this is incremented by the 
		             appropriate value for each instruction; 
			       GLOBAL VARIABLE */

   int  lineno;       /* line number; this is incremented by 1 for each input 
		             assembly language line;  
			       GLOBAL VARIABLE */


   struct errtbl eque;  /* Error Queue; all the errors for the input 
			         are maintained in a queue; 
			         GLOBAL VARIABLE */

   errnode enode, *enodeptr; /* Node and pointer to node in the Error Queue */

   struct opndque opque; /* Operand Queue; all the labels in the expression 
	        	          are maintained in a queue; 
			          GLOBAL VARIABLE */    

   opqnode onode, *onodeptr;/* Node and pointer to node in the Operand Queue*/

/*********************     P R O T O T Y P E S     **************************/

/*-----------------------    OPERATION TABLE    ----------------------------*/

   void optab_bubblesort(struct ot_entry *, int);
   /* This routine is used by the pass1_init routine.  
      It sorts the values in optab.*/

   void optab_swap(struct ot_entry *, struct ot_entry *);
   /* This routine is used by optab_bubblesort. 
      It swaps two values in the optab. */

   void optab_write(FILE **, struct ot_entry *, int);
   /* This routine is used by the pass1_init routine.  
      It writes the optab to the ouput file. */

   int optab_search(char *, struct ot_entry *, int);
   /* Binary search on the sorted Optab; 
      takes this operator, the optab and the size of the table; 
      returns index to location if found or NADA (-1) if not found.  */

/*------------------------     SYMBOL TABLE    -----------------------------*/
   void symtab_init(struct st_entry *);
   /* Initializes the symbol table. */

   int symtab_hash(char *);
   /* Takes a pointer to a char array and returns the sum of the first and 
      last capital letters (in the char array) - 2, where A = 1 and Z = 26.
      For example CONE returns 3 + 5 - 2 = 6. */

   int symtab_search(char *, struct st_entry *, int *);
   /* Takes the label, symtab, and a pionter to the original hash index; 
      searches the symtab for label - if found returns TRUE; 
      else returns false; the index becomes the latest place to insert.

   void symtab_insert(char *, char , struct st_entry *, int );
   /* Pre-condition: symtab is not full and label is not found;
      Post-condition: struct of label name, type and value is inserted 
      into the symtab.  
      Takes the label, type, symtab and the index (already hashed and 
      searched) to the empty slot. */

   void symtab_write(struct st_entry *, FILE **);
   /* Writes the symbol table to the output file. */

/*-------------------------     ERROR QUEUE    -----------------------------*/

   void errque_init(struct errtbl *);
   /* This routine simply initializes the error queue. */

   void errque_insert(struct errtbl *, int);
   /* This routine allows us add new errors into our error queue by 
      malloc-ing space and filling the new node with pertinent information, 
      then enqueing that node into our queue. 
      Takes the error queue and the index value corresponding to the type 
      of error encountered. */

   void errque_write(struct errtbl *, FILE **);
   /*  Writes the error queue to the output file.*/

/*------------------------     OPERAND QUEUE    ----------------------------*/

   void opndque_init(struct opndque *);
   /* This routine simply initializes the operand queue. */

   void opndque_insert(struct opndque *, opqnode *);
   /* This routine inserts an operand node into the operand queue
      Takes the operand queue and the already malloced and filled node 
      (the node gets created and filled in the function, parse_operand). */

   void opndque_evaluate(struct opndque *, int *);
   /* Evaluates the expression in the operand queue, passing it back 
      through the integer arguement */

   void opndque_flush(struct opndque *opq);
   /* Flushes the operand queue freeing the memory; does not initialize. */

/*-----------------------------   MAIN   -----------------------------------*/

   void pass1_init(int, char **);
   /* This routine is used by the main routine.
      It takes the numver of arguements and the argv array of argurments.
      It initializes the global variables locctr and lineno.
      It opens the three files (sets ot_fptr, in_fptr, out_fptr).
      It loads optab (using ot_fptr) and sorts optab.
      It writes the sorted optab to output file.
      It initializes symtab.
      It initializes errque.*/


   void pass1(void);
   /* This routine is used by the main. It performs Pass1 of the assembler. */

   void get_tokens(char *, char *, char *, char *);
   /* This routine is used by pass1;  takes the line, label, operator, and 
      operand - parses the line and fills the three variables respectively. */

   void machine_processing(FILE **fp, struct st_entry *st,
                     char lbl[NAME_LENGTH + 1], char opr[OP_LENGTH + 1],
                     char ond[OPND_LENGTH + 1], int op_idx);
   /* This routine is used by pass1 to process machine operations. 
      It takes a pointer to the out file pointer, the symbol table, the label, 
      the operator, the operand and the operator table index */

   void dir_processing(FILE **fp, struct st_entry *st, struct opndque *opq, 
                     char lbl[NAME_LENGTH + 1], char opr[OP_LENGTH + 1], 
                     char ond[OPND_LENGTH + 1], int op_idx);
   /* This routine is used by pass1 to process assembler directives. 
      It takes a pointer to the out file pointer, the symbol table, the 
      operand queue, the label, the operator, the operand and the 
      operator table index */

   int parse_operand (char *, struct opndque *, struct st_entry *, char *);
   /* This routine is used by dir_processing when an EQU is the operator. 
      It takes the operand expression which has been parsed from inp_line,
      the operand queue, the symbol table and a tempvariable to hold the type.
      It returns true if the operand expression is valid and false if it is 
      not.  The type ('R' or 'A' has been passed back through the last 
      argument.*/

   void print_info(struct errtbl *, struct st_entry *);
   /* This routine simply prints out the error queue and the symbol table. */


/***************************        M A I N       ****************************/

   int main (int argc, char **argv)
   {
      pass1_init(argc, argv);
   
      pass1();
   
      print_info(&eque, symtbl);
   
      printf("\nPass1 has successfully processed your assembly ");
      printf("program, \"%s\".",  argv[2]);
      printf("\n\nThe assembly listing and other vital ");
      printf("data have been directed");
      printf("\nto your indicated ouput file, \"%s\".\n", argv[3]);
   
      fclose(ot_fptr);
      fclose(in_fptr);
      fclose(out_fptr);
   
      return EXIT_SUCCESS;
   
   } /* end MAIN */

/***************************   F U N C T I O N S  ****************************/
/*-----------------------    OPERATION TABLE    ----------------------------*/

   void optab_bubblesort(struct ot_entry ot[], int n)
   /* This routine is used by the pass1_init routine.  
      It sorts the values in optab.*/
   {
      int i, j;
   
      for (i=0; i< n-1; ++i)
         for (j=n-1; i<j; --j)
            if(strcmp(ot[j-1].ot_mnemonic, ot[j].ot_mnemonic)>0)
               optab_swap(&ot[j-1], &ot[j]);
   
   } /* end Bubble */

   void optab_swap(struct ot_entry *one, struct ot_entry *two)
   /* This routine is used by optab_bubblesort. 
      It swaps two values in the optab. */
   {
      struct ot_entry temp;
   
      temp= *one;
      *one=*two;
      *two= temp;
   
   } /* end optab_swap */

   void optab_write(FILE **fp, struct ot_entry *ot, int ot_cnt)
   /* This routine is used by the pass1_init routine.  
      It writes the optab to the ouput file. */
   {
      int j;
      fprintf(*fp, "   The Operation Table\n");
      fprintf(*fp, "Index   Operation   Format\n");
      for (j = 0; j < ot_cnt; j++)
      {
         fprintf(*fp,"%3d     %-8s    %3d\n",
                j, ot[j].ot_mnemonic, ot[j].ot_format);
      }  /* end for */
   
   } /* end write_optab */

   int optab_search(char op[OP_LENGTH + 1], struct ot_entry ot[], int s)
   /* Binary search on the sorted Optab; 
      takes this operator, the optab and the size of the table; 
      returns index to location if found or NADA (-1) if not found.  */
   {
      int condition,
      low, high, mid;
   
      low = 0;
      high = s - 1;
      while (low <= high)
      {
         mid = (low + high) / 2;
         if ((condition = strcmp(op, ot[mid].ot_mnemonic)) < 0)
            high = mid - 1;
         else if (condition > 0)
            low = mid +1;
         else
            return mid;
      } /*end while */
   
      return NADA;
   
   } /* end optab_seach */

/*------------------------     SYMBOL TABLE    -----------------------------*/

   void symtab_init(struct st_entry *st)
   /* Initializes the symbol table. */
   {
      int i;
      for (i=0; i< HT_SIZE; i++)
      {
         st[i].st_name[0]='\0';
         st[i].st_type='Z';
         st[i].st_value=NADA;
      }
   } /* end symtab_init */

   int symtab_hash(char *lbl_ptr)
   /* Takes a pointer to a char array and returns the sum of the first and 
      last capital letters (in the char array) - 2, where A = 1 and Z = 26.
      For example CONE returns 3 + 5 - 2 = 6. */
   {
      char first_cap = 0;
      char last_cap = 0;
      while (*lbl_ptr != '\0'){
         if (*lbl_ptr >= 'A' && *lbl_ptr <= 'Z'){
            if (!first_cap){
               first_cap = last_cap = *lbl_ptr - 'A' + 1;
            }
            else {
               last_cap = *lbl_ptr - 'A' + 1;
            }
         }
         lbl_ptr++;
      }
   
      return (first_cap + last_cap) - 2;
   
   } /* end symtab_hash() */

   int symtab_search(char lbl[NAME_LENGTH + 1], struct st_entry *st, int *sidx)
   /* Takes the label, symtab, and a pionter to the original hash index; 
      searches the symtab for label - if found returns TRUE; 
      else returns false; the index becomes the latest place to insert. */
   {
      int found_flag= FALSE;
      int empty_flag= FALSE;
      int condition;
      int c=0;
   
      while (!(empty_flag) && !(found_flag) && (c < HT_SIZE))
      {
         condition = strcmp ((st + *sidx)->st_name, lbl);
         c++;
         if (condition == 0)
         {
            found_flag=TRUE;
         }
         else if (*(st + *sidx)->st_name == '\0')
         {
            empty_flag= TRUE;
         }
         else
         {
            *sidx = (*sidx + STEP_SIZE) % HT_SIZE;
            c++;
         }
      }
      if ((c >= HT_SIZE) && (found_flag == FALSE))
      {
         fprintf(stderr, "Error - Symbol Table Overflow");
         exit(1);
      }
   
      return found_flag;
   
   } /*end search */

   void symtab_insert(char lbl[NAME_LENGTH + 1], char lbl_type,
                     struct st_entry *st, int sidx)
   /* Pre-condition: symtab is not full and label is not found;
      Post-condition: struct of label name, type and value is inserted 
      into the symtab.  
      Takes the label, type, symtab and the index (already hashed and 
   	searched) to the empty slot. */
   {
      strcpy(st[sidx].st_name, lbl);
      st[sidx].st_type=lbl_type;
      st[sidx].st_value=locctr;
   
   } /* end symtab_insert */

   void symtab_write(struct st_entry *st, FILE **fp)
   /* Writes the symbol table to the output file. */
   {
      int j = 0;
      fprintf(*fp, "\n        The Symbol Table\n");
      fprintf(*fp, "Hash Index   Name          Value   Type\n");
      while(j < HT_SIZE)
      {
         if  (st[j].st_name[0] != '\0')
         {
            fprintf(*fp,"%5d        %-13s%-9.4X%c\n",
                   j, st[j].st_name, st[j].st_value, st[j].st_type);
         } /* end if */
         j++;
      }  /* end while */
   
   } /* end symtab_write */

/*-------------------------     ERROR QUEUE    -----------------------------*/

   void errque_init(struct errtbl *eq)
   /* This routine simply initializes the error queue. */
   {
      eq->front=NULL;
      eq->rear=NULL;
   
   }/* end errque_init */


   void errque_insert(struct errtbl *eq, int i)
   /* This routine allows us add new errors into our error queue by 
      malloc-ing space and filling the new node with pertinent information, 
      then enqueing that node into our queue. 
      Takes the error queue and the index value corresponding to the type 
      of error encountered. */
   {
      errnode  *eptr_temp, *tracker;
   
      tracker = eq->front;
      eptr_temp= ((errnode *)malloc(sizeof(errnode)));
      eptr_temp->eq_lineno= lineno;
      eptr_temp->eq_msgind= i;
      eptr_temp->eq_next= NULL;
   
      if (eq->front==NULL)
      {
         eq->front=eptr_temp;
         eq->rear= eptr_temp;
      }
      else
      {
         while (tracker->eq_next != NULL)
            tracker = tracker->eq_next;
         tracker->eq_next = eptr_temp;
         eq->rear = eptr_temp;
      }
   } /* end errque_insert */


   void errque_write(struct errtbl *eq, FILE **fp)
   /*  Writes the error queue to the output file.*/
   {
      errnode *tracker;
   
      tracker = eq->front;
      fprintf(*fp,"\nError Table\n");
      fprintf(*fp,"Line   Error\n");
      fprintf(*fp,"%3d    %s\n",tracker->eq_lineno, ERR_MSGS[tracker->eq_msgind]);
   
      while (tracker !=eq->rear)
      {
         tracker= tracker->eq_next;
         fprintf(*fp,"%3d    %s\n",
                tracker->eq_lineno, ERR_MSGS[tracker->eq_msgind]);
      }
   } /* end errque_write */


/*------------------------     OPERAND QUEUE    ----------------------------*/

   void opndque_init(struct opndque *opq)
   /* This routine simply initializes the operand queue. */
   {
      opq->qlength = 0;
      opq->front=NULL;
      opq->rear=NULL;
   
   } /* end opndque_init */

   void opndque_insert(struct opndque *opq, opqnode *opn)
   /* This routine inserts an operand node into the operand queue
      Takes the operand queue and the already malloced and filled node 
      (the node gets created and filled in the function, parse_operand). */
   {
      opqnode   *tracker;
   
      tracker = opq->front;
   
      if (opq->front==NULL)
      {
         opq->qlength++;
         opq->front=opn;
         opq->rear= opn;
      }
      else
      {
         while (tracker->opq_next != NULL)
         {
            tracker = tracker->opq_next;
         }; 
         tracker->opq_next = opn;
         opq->rear = opn;
         opq->qlength++;
      }
   } /* end opndque_insert */


   void opndque_evaluate(struct opndque *opq, int *tvalue)
   /* Evaluates the expression in the operand queue, passing it back 
      through the integer arguement */
   {
      int i;
      opqnode *tracker;
   
      tracker = opq->front;
      for (i=1; i<=opq->qlength; i++)
      {
         if (tracker->opq_sign == PLUS)
            *tvalue += (tracker->opq_value);
         if (tracker->opq_sign == MINUS)
            *tvalue -= (tracker->opq_value);
         tracker= tracker->opq_next;
      }
   } /* end opndque_evaluate */


   void opndque_flush(struct opndque *opq)
   /* Flushes the operand queue freeing the memory; does not initialize. */
   {
      opqnode *tempptr;
      tempptr =  opq->front;
   
      while (tempptr->opq_next != NULL)
      {
         opq->front = tempptr -> opq_next;
         free(tempptr);
         tempptr =opq->front;
      } /* end while */
      free(tempptr);
   
      return;
   } /* end opndque_flush */

/*-----------------------------   MAIN   -----------------------------------*/

   void pass1_init(int argc, char *argv[])
   /* This routine is used by the main routine.
      It takes the numver of arguements and the argv array of argurments.
      It initializes the global variables locctr and lineno.
      It opens the three files (sets ot_fptr, in_fptr, out_fptr).
      It loads optab (using ot_fptr) and sorts optab.
      It writes the sorted optab to output file.
      It initializes symtab.
      It initializes errque.*/
   
   { /* begin pass1_init */
      locctr=0;
      lineno=0;
      ot_count=0;
   
      if (argc < 4)
      {
         printf("You did not enter the correct number of file arguements.");
         exit(1);
      }; /* end if */
   
      if((ot_fptr=fopen(argv[1], "r")) == NULL) /* opens argv[1] file */
      {
         fprintf(stderr, "Error opening optab file.");
         exit(1);
      };
      if((in_fptr=fopen(argv[2], "r")) == NULL)   /* opens argv[2] file */
      {
         fprintf(stderr, "Error opening input file.");
         exit(1);
      };
      if( ( out_fptr=fopen(argv[3], "w")) == NULL)   /* opens argv[3] file */
      {
         fprintf(stderr, "Error opening output file.");
         exit(1);
      };
      symtab_init(symtbl);
      errque_init(&eque);
      opndque_init(&opque);
   
   /*---------------     Creating the Op table      ------------------*/
      while (fscanf(ot_fptr, "%s", optbl[ot_count].ot_mnemonic)!=EOF)
      {
         fscanf(ot_fptr, "%d", &optbl[ot_count].ot_format);
         ot_count++;
      }
   
      optab_bubblesort(optbl, ot_count);
      optab_write(&out_fptr, optbl, ot_count);
   
   } /* end pass1_init */

   void pass1(void)
   /* This routine is used by the main. It performs Pass1 of the assembler. */
   {
      char inp_line[INP_LENGTH + 1]; /* input assembly line */
      char new_oper[OP_LENGTH + 1];  /* Temporary space */
      char label[NAME_LENGTH +1];    /* label in the input */
      char oper[OP_LENGTH + 1];      /* operation in the input */
      char opnd[OPND_LENGTH + 1];    /* operand in the input; */
                                /*this will contain everything after the
                                  operation in the input*/
   
      int op_index;
   
      /* read each input assembly line and perform pass-1 */
      fprintf(out_fptr, "\n     The Assembly Line Listing    \n");
      fprintf(out_fptr, "Line   Loc    Source Statement");
      fprintf(out_fptr,"                     Error   Op Index\n");
   
      while (fgets(&inp_line[0], INP_LENGTH, in_fptr) != NULL)
      {
         if (inp_line[0] != '\n')
            lineno++;  /* This may be nit-picky, but it allows for assembly 
                          code with double spacing between lines. */
      
         if (inp_line[0] != '.' && inp_line[0] !='\n')
         /*As long as the line to be processed is NOT a blank line,
           NOR is it an assembler directive then it will be parsed and
           written out to out.txt */
         {
            get_tokens(inp_line, &label[0], &oper[0], &opnd[0]);
         }
         else
         {
            if (inp_line[0] !='\n' && inp_line[0]=='.')
            {
               label[0]='\0';
               oper[0]='\0';
               opnd[0]='\0';
               fprintf(out_fptr, "%3d           %s",lineno, inp_line);
            }
         }
      
      /*----------------      search optab      -----------------------*/
         if (inp_line[0] != '.' && inp_line[0] !='\n')
         {
            if (oper[0]=='+')
            {
               strcpy(new_oper, oper+1);
               op_index= optab_search(new_oper, optbl, ot_count);
            }
            else
               op_index=optab_search(oper, optbl, ot_count);
         
            if (op_index == NADA)  /* If operator NOT defined in your optab then error!*/
            {
               errque_insert(&eque, INV_OP);
               fprintf(out_fptr, "%3d%22.11s%-7s%-21sY\n",
                      lineno,label, oper, opnd);
            }
            else if(optbl[op_index].ot_format == 0)
            {
               dir_processing(&out_fptr, symtbl, &opque, label, oper, opnd,
                             op_index);
            }
            /*  machine_processing;*/
            
            else
            {
               machine_processing(&out_fptr, symtbl, label, oper, opnd,op_index);
            
               if (oper[0]=='+')
                  locctr+=4;
               else
                  locctr+= optbl[op_index].ot_format;
            }
         
         }/* end if */
      } /* end while get line is not null */
   
   } /* end pass1 */

   void get_tokens(char line[INP_LENGTH +1],char lbl[NAME_LENGTH +1],
                  char opr[OP_LENGTH + 1], char ond[OPND_LENGTH + 1])
   /* This routine is used by pass1;  takes the line, label, operator, and 
      operand - parses the line and fills the three variables respectively. */
   {
   
      char *p;
   
      lbl[0]='\0';
      opr[0]='\0';
      ond[0]='\0';
   
      if (line[0] != '\n')
      {
         if ((line[0] != ' ') && (line[0] != '\t')
            && (line[0] != '\0'))
         {
            if ((p = strtok(line, DELIM)) != NULL)
               strcpy(lbl, p);
            if ((p = strtok(NULL, DELIM)) != NULL)
               strcpy(opr, p);
            if ((p = strtok(NULL, DELIM)) != NULL)
               strcpy(ond, p);
         }
         else
         {
            if ((p = strtok(line, DELIM)) != NULL)
               strcpy(opr, p);
            if ((p = strtok(NULL, DELIM)) != NULL)
               strcpy(ond, p);
         }
      }
      #ifdef DEBUG
         printf("%-11s%-7s%-7s\n", lbl, opr, ond);
      #endif
   
   } /* end get_tokens */

   void machine_processing(FILE **fp, struct st_entry *st,
                     char lbl[NAME_LENGTH + 1],
                     char opr[OP_LENGTH + 1],
                     char ond[OPND_LENGTH + 1],
                     int op_idx)
   
   /* This routine is used by pass1 to process machine operations. 
      It takes a pointer to the out file pointer, the symbol table, the label, 
      the operator, the operand and the operator table index */
   {
      int st_index;
   
   
      if (lbl[0] != '\0')
      {
      /* This if statement will process a label--first making sure it has not been
      defined already. Provided the table is not already full, then it will insert
      the label at appropriate hash value, resolving collisions if necesscary */
         st_index= symtab_hash(lbl);
      
         if (symtab_search(lbl, st, &st_index))
         {
            errque_insert(&eque, DUP_LBL);
            fprintf(*fp, "%3d           %-11s%-7s%-21sY\n",
                   lineno, lbl, opr, ond);
         }
         else
         {
            symtab_insert(lbl, RELOCATABLE, st, st_index);
            fprintf(*fp, "%3d    %-7.4X%-11s%-7s%-21s%11d\n",
                   lineno,locctr,lbl, opr, ond, op_idx);
         }
      }/* end if */
      else
      {
         fprintf(*fp, "%3d    %-7.4X%-11s%-7s%-21s%11d\n",
                lineno,locctr,lbl, opr, ond, op_idx);
      }
   } /* end machine_processing */


   void dir_processing(FILE **fp, struct st_entry *st, struct opndque *opq,
                     char lbl[NAME_LENGTH + 1], char opr[OP_LENGTH + 1],
                     char ond[OPND_LENGTH + 1],	int op_idx)
   /* This routine is used by pass1 to process assembler directives. 
      It takes a pointer to the out file pointer, the symbol table, the 
      operand queue, the label, the operator, the operand and the 
      operator table index */
   {
      char temp_type;
      int  valid_expression = TRUE;
      int  temp_value = 0;
      int  op_index;
      int  st_index;
   /*      int  ond_index;   */
      int  temp_int;
   
      if (strcmp(opr, "START")==0 || strcmp(opr, "END")==0)
      {
         if (strcmp(opr, "START")==0)
         {
            st_index= symtab_hash(lbl);
            if (strcmp(opr, "END")!=0 && symtab_search(lbl, st, &st_index))
            {
               errque_insert(&eque, DUP_LBL);
               fprintf(*fp, "%3d           %-11s%-7s%-21sY\n",
                      lineno, lbl, opr, ond);
            }
            else
            {
               symtab_insert(lbl, RELOCATABLE, st, st_index);
               fprintf(*fp, "%3d    %-7.4X%-11s%-7s%-21s%11d\n",
                      lineno,locctr,lbl, opr, ond, op_idx);
            }
         } /*end if start */
         else if (strcmp(opr, "END")==0)
         {
            st_index= symtab_hash(lbl);
            if (strcmp(opr, "END")!=0 && symtab_search(lbl, st, &st_index))
            {
               errque_insert(&eque, DUP_LBL);
               fprintf(*fp, "%3d           %-11s%-7s%-21sY\n",
                      lineno, lbl, opr, ond);
            }
            else
            {  /* if operand is not a valid label - halt program */
               op_index= symtab_hash(ond);
               if (!symtab_search(ond, st, &op_index))
               {
                  fprintf(stderr, "Error: END refers to an invalid label");
                  exit(1);
               }
               else
               {  /* if LABEL insert into symtab */
                  if (lbl[0] != '\0')
                     symtab_insert(lbl, RELOCATABLE, st, st_index);
                  fprintf(*fp, "%3d    %-7.4X%-11s%-7s%-21s%11d\n",
                         lineno,locctr,lbl, opr, ond, op_idx);
               }
            }
         } /*end if start */
      
      }
      
      else if (lbl[0] != '\0' && strcmp(opr, "EQU")!=0)
      {
         st_index= symtab_hash(lbl);
         if (symtab_search(lbl, st, &st_index))
         {
            errque_insert(&eque, DUP_LBL);
            fprintf(*fp, "%3d           %-11s%-7s%-21sY\n",
                   lineno, lbl, opr, ond);
         }
         else
         {
            fprintf(*fp, "%3d    %-7.4X%-11s%-7s%-21s%11d\n",
                   lineno,locctr,lbl, opr, ond, op_idx);
            if (strcmp(opr,"RESB")==0)
            {
               temp_int =atoi(ond); /* Note: atoi is a function that converts an */
               symtab_insert(lbl, RELOCATABLE, st, st_index);
               locctr+= temp_int;   /* array of chars (a string) to an integer */
            }
            if (strcmp(opr,"RESW")==0)
            {
               temp_int= atoi(ond);
               symtab_insert(lbl, RELOCATABLE, st, st_index);
               locctr+= (temp_int*3);
            }
            if (strcmp(opr, "BYTE")==0)
            {
               if (ond[0]=='C')
               {
                  temp_int= (strlen(ond)-3);
                   /* Subtract 3 because for example C'1300' is 7
                      chars long. But only four that we care about*/
                  symtab_insert(lbl, RELOCATABLE, st, st_index);
                  locctr+=temp_int;
               }
               else if (ond[0]=='X')
               {
                  temp_int= (strlen(ond)-3);
                  temp_int= (temp_int+1)/2;
                  locctr+=temp_int;
               }
            } /* End if opr==BYTE */
         
         }
      
      } /* End BIG else if */
   
      if (strcmp(opr,"EQU")==0)
      {
         st_index= symtab_hash(lbl);
         if (symtab_search(lbl, st, &st_index))
         {
            errque_insert(&eque, DUP_LBL);
            fprintf(*fp, "%3d           %-11s%-7s%-21sY\n", lineno, lbl, opr, ond);
         }
         if (strcmp(ond,"*") != 0)
         {
            if (atoi(ond)!=0 && !isalpha(ond[0]))
             /* Provided the operand value is NOT an integer and the first part
                of the string in the operand is NOT a letter then we conclude that
                the operand is a number and we can equate the label in the label
                field to the integer in the operand */
            {
               symtab_insert(lbl, ABSOLUTE, st, st_index);
               st[st_index].st_value= atoi(ond);
            }
            else                                             
            {
            /* If the operand can be parsed into labels it will be done here.
               op_index1 and 2 store the index value of the operaters. The operator
               is then inserted into symtab, and finally it's st_value is changed*/
               valid_expression = parse_operand (ond, &opque, st, &temp_type);
               if (valid_expression)
               {
                  opndque_evaluate(opq, &temp_value);
                  symtab_insert(lbl, temp_type, st, st_index);
                  st[st_index].st_value = temp_value;
                  opndque_flush(opq);
                  opndque_init(opq);
               }
               else
               {
                  fprintf(stderr, "Error: you programmed an invalid expression");
                  exit(1);
               }
            
            }
            if(strlen(ond) > 21)
            {
               fprintf(*fp, "%3d    %-7.4X%-11s%-7s%-29s%3d\n",
                      lineno,locctr,lbl, opr, ond, op_idx);
            }
            else
               fprintf(*fp, "%3d    %-7.4X%-11s%-7s%-21s%11d\n",
                      lineno,locctr,lbl, opr, ond, op_idx);
         }
         if (strcmp(ond,"*")==0)
         {
            symtab_insert(lbl, RELOCATABLE, st, st_index);
            st[st_index].st_value=locctr;
            fprintf(*fp, "%3d    %-7.4X%-11s%-7s%-21s%11d\n",
                   lineno,locctr,lbl, opr, ond, op_idx);
         }
      } /* End if opr==EQU */
   
   
   } /* end dir_processing */

   int parse_operand (char ond[OPND_LENGTH+1], struct opndque *opq,
                     struct st_entry *st, char *otype)
   /* This routine is used by dir_processing when an EQU is the operator. 
      It takes the operand expression which has been parsed from inp_line,
      the operand queue, the symbol table and a tempvariable to hold the type.
      It returns true if the operand expression is valid and false if it is 
      not.  The type ('R' or 'A' has been passed back through the last 
      argument.*/
   
   {
      int  start = 0,
      i = 0,
      valid = TRUE,		/* if expressin is valid */	
      pos = 0,    		/* to count the positives */
      neg = 0,    		/* to count the negatives */
      r_count = 0, 	 	/* to keep track of relocatables */
      ondex,        		/* hash index of operand */
      strl;         		/* length of entire onperand expression */
      char temp[NAME_LENGTH +1];
      opqnode *optr_temp;     /* pointer to node for operand queue */
      temp[0]='\0';
   
      strl = strlen(ond);
   
      if (ond[i] != '\n')   /* Brute force methond of parsing...*/
      {
         while (i < strl)
         {
            if(ond[i] == MINUS)
            {
               ++i;
               while (ond[i] != MINUS && ond[i] != PLUS && ond[i] != '\0')
               {
                  temp[start] = ond[i];
                  ++i;
                  start++;
               }  /* end while */
               temp[start] = '\0';
               start = 0;
               optr_temp= ((opqnode *)malloc(sizeof(opqnode)));
               optr_temp->opq_next = NULL;
               strcpy(optr_temp->opq_name, temp);
               optr_temp->opq_sign = MINUS;
               if (isalpha(temp[0]))
               {
                  ondex = symtab_hash(temp);
                  if (symtab_search(temp, st, &ondex))
                  {
                     optr_temp->opq_type = st[ondex].st_type;
                     optr_temp->opq_value = st[ondex].st_value;
                     if (optr_temp->opq_type == RELOCATABLE)
                     {
                        neg++;
                        r_count++;
                     }
                  }
                  else
                  {
                     fprintf(stderr, "Error - Operand Expression not in Symtab");
                     exit(1);
                  }
               } /* end if label */
               else  /*integer*/
               {
                  optr_temp->opq_type = ABSOLUTE;
                  optr_temp->opq_value = atoi(temp);
               }
               opndque_insert(opq, optr_temp);
            
            }/* end if neg */
            else
            {
               if (ond[i] == PLUS)
                  i++;
               while (ond[i] != MINUS && ond[i] != PLUS && ond[i] != '\0')
               {
                  temp[start] = ond[i];
                  ++i;
                  start++;
               }  /* end while */
               temp[start] = '\0';
               start = 0;
               optr_temp= ((opqnode *)malloc(sizeof(opqnode)));
               optr_temp->opq_next = NULL;               
               strcpy(optr_temp->opq_name, temp);
               optr_temp->opq_sign = PLUS;
               if (isalpha(temp[0]))
               {
                  ondex = symtab_hash(temp);
                  if (symtab_search(temp, st, &ondex))
                  {
                     optr_temp->opq_type = st[ondex].st_type;
                     optr_temp->opq_value = st[ondex].st_value;
                     if (optr_temp->opq_type == RELOCATABLE)
                     {
                        pos++;
                        r_count++;
                     }
                  }
                  else
                  {
                     fprintf(stderr, "Error - Operand Expression not in Symtab");
                     exit(1);
                  
                  }
               } /* end if label */
               else  /*integer*/
               {
                  optr_temp->opq_type = ABSOLUTE;
                  optr_temp->opq_value = atoi(temp);
               }
               opndque_insert(opq, optr_temp);
            }/* end else pos */
         }/* end while  != '\0' */
      } /*end if */
      if ((pos - neg) == 0 || (pos - neg) == 1)
         valid = TRUE;
      else
         valid = FALSE;
      if (valid && ((r_count%2) == 1))
      {
         *otype = RELOCATABLE;
      }
      else if (valid && ((r_count%2) == 0))
      {
         *otype = ABSOLUTE;
      }
      return valid;
   
   } /*End parse_operand */

   void print_info(struct errtbl *eq, struct st_entry *st)
   /* This routine simply prints out the error queue and the symbol table. */
   {
      errque_write(eq, &out_fptr);
      symtab_write(st, &out_fptr);
   
   } /* end print_info */




